##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import numpy as _np
from . import _utils as _futils
from mdciao.plots.plots import _points2dataunits
from mdciao.utils.str_and_dict import replace4latex
from mdciao.utils.residue_and_atom import get_SS as _get_SS
from mpl_chord_diagram import chord_diagram as _chord_diagram
from collections import namedtuple as _namedtuple

from matplotlib import pyplot as _plt
from matplotlib.lines import Line2D as _Line2D
from matplotlib.patches import CirclePolygon as _CP
from inspect import signature as _signature

def freqs2flare(freqs, res_idxs_pairs,
                fragments=None,
                sparse_residues=False,
                sparse_fragments=False,
                exclude_neighbors=1,
                freq_cutoff=0,
                iax=None,
                fragment_names=None,
                center=_np.array([0,0]),
                r=1,
                mute_fragments=None,
                anchor_fragments=None,
                SS=None,
                panelsize=10,
                angle_offset=0,
                highlight_residxs=None,
                select_residxs=None,
                top=None,
                colors=True,
                fontsize=None,
                shortenAAs=True,
                aa_offset=0,
                markersize=None,
                bezier_linecolor='k',
                plot_curves_only=False,
                textlabels=True,
                padding=[1,1,1],
                lw=5,
                signed_colors=None,
                subplot=False,
                aura=None,
                coarse_grain=False
                ):
    r"""
    Plot contact frequencies as `flare plots`.

    The residues are plotted as dots lying on a circle of radius
    :obj:`r`, with Bezier curves of variable opacity connecting
    the dots. The curve opacity represents the contact frequency, :obj:`freq`,
    between two residues.

    For more info see the links here :obj:`mdciao.flare`.

    One can control separately what residues and what curves
    ultimately get shown, allowing for "contactless" residues
    to still appear as dots in the circle. This is very helpful
    to highlight the molecular topology. For example, it's useful to
    show an entire TM-helix even if only the residues in its middle
    have contacts. Furthermore, it allows for re-use of a "background"
    of residues on top of which different sets of curves (e.g. with
    different colors) can be plotted.

    * which/how residues get plotted:
        * :obj:`res_idxs_pairs` is the primary source of which
          residues will be plotted, its length matches the
          length the :obj:`freqs` object.
        * :obj:`fragments` is used to
          a) expand the initial residue list (if needed) and
          b) split the residues into fragments when placing the dots on the flareplot.
        * :obj:`sparse_residues` can modify :obj:`res_idxs_pairs`,
          either by looking at non-zero :obj:`freqs` or
          by direct input (see below)
        * :obj:`sparse_fragments` is analogous, but with
          fragments
        * :obj:`highlight_residxs` show the labels of these residues
          in red

    * which/how residues get connected with bezier curves:
        * :obj:`exclude_neighbors` do not plot curves connecting neighbors
        * :obj:`freq_cutoff` do not plot curves with associated :obj:`freq`
          below this cutoff
        * :obj:`select_residxs` plot only the curves where these residues appear
        * :obj:`mute_fragments` do not plot curves from/to this fragment
        * :obj:`anchor_fragments` only plot curves from/to this fragment

    Parameters
    ----------
    freqs : numpy.ndarray
        The contact frequency to show in the flareplot. The
        opacity of the bezier curves connecting residue
        pairs will be proportional to this number. Can
        have different shapes:

        * (n)
          n is the number of residue pairs in :obj:`res_idxs_pairs`
        * (m,n)
          m is the number of frames, in this case,
          an average over m will be done automatically.

        If you already have an :obj:`mdciao.contacts.ContactGroup`-object,
        simply pass the result of calling
        :obj:`mdciao.contacts.ContactGroup.frequency_per_contact` (with
        a given cutoff)
    res_idxs_pairs : iterable of pairs
        pairs of residue indices for the above N contact frequencies.
        If you already have an :obj:`mdciao.contacts.ContactGroup`-object,
        simply pass the result of calling
        :obj:`mdciao.contacts.ContactGroup.res_idxs_pairs`
    fragments: list of lists of integers, default is None
        The residue indices to be drawn as a circle for the
        flareplot. These are the dots that will be plotted
        on that circle regardless of how many contacts they
        appear in. They can be any integers that could represent
        a residue. The only hard condition is that the set
        of np.unique(res_idxs_pairs) must be contained
        within :obj:`fragments`
    sparse_residues : boolean, default is False
        Show only those residues that appear in the initial :obj:`res_idxs_pairs`

        Note
        ----
        There is a development option for this argument where a residue
        list is passed, meaning, show these residues regardless of any other
        option that has been passed. Perhaps this changes in the future.
    sparse_fragments : boolean, default is False
        Same as :obj:`sparse_residues`, but with fragments. When
        :obj:`sparse_residues` isn't False, this option
        has no effect.
    exclude_neighbors: int, default is 1
        Do not show contacts where the partners are separated by
        these many residues. If no :obj:`top` is passed, the
        neighborhood-condition is checked using residue
        serial-numbers, assuming the molecule only has one long peptidic-chain.
    freq_cutoff : float, default is 0
        Contact frequencies lower than this value will not be shown
    iax : :obj:`~matplotlib.axes.Axes`
        Parse an axis to draw on, otherwise one will be created
        using :obj:`panelsize`. In case you want to
        re-use the same cirlce of residues as a
        background to plot different sets
        of :obj:`freqs`, **YOU HAVE TO USE THE SAME**
        :obj:`fragments` and :obj:`sparse` values
         **on all calls**, else the
        bezier lines will be placed erroneously.
    fragment_names: iterable of strings, default is None
        The names of the fragments used in :obj:`fragments`
    center: np.ndarray, default is [0,0]
        In axis units, where the flareplot will be centered around
    r: float, default is 1
        In axis units, the radius of the flareplot
    mute_fragments: iterable of integers, default is None
        Curves involving these fragments will be hidden. Fragments
        are expressed as indices of :obj:`fragments`
    anchor_fragments: iterable of integers, default is None
        Curves **not** involving these fragments will be
        **not** be shown, i.e. it is the complementary
         of :obj:`mute_fragments`. Both cannot be passed
         simultaneously.
    SS : any, default is None
        Can be several things:
        * Array containing secondary structure (ss) information to
          be included in the flareplot. Indexed by residue index,
          i.e. it can also be a dictionary as long as
          SS[idx] returns the SS for residue with that residue idx
        * Path to filename, will be passed to
        :obj:`mdciao.utils.residue_and_atom.get_SS`, check
        the docs there
    panelsize: float, default is 10
        Size in inches of the panel (=figsize in matplotlib).
        Will be ignored if a pre-existing axis object is parsed
    angle_offset : float, default is 0
        In degrees, where the flareplot "begins". Zero is xy = [1,0]
    highlight_residxs : iterable of ints, default is None
        Show the labels for these residues in red
    select_residxs : iterable of ints, default is None
        Only the residues here can be connected with a Bezier curve
    top: :obj:`~mdtraj.Topology` object, default is None
        If provided a top, residue names (e.g. GLU30) will be used
        instead of residue indices. Will fail if the residue indices
        in :obj:`res_idxs_pairs` can not be used to call :obj:`top.residue(ii)`
    colors: boolean, default is True
        Color control. Can take different inputs
         * True: use one different color per segment
         * False: defaults to gray
         * str or char: use that color for all residues (e.g. "r" or "red")
         * A list of strings of len = number of drawn residues, which is
           equal to len(np.hstack(fragments))
         * A list of strings of len = len(fragments)
        Any other length will produce an error
    fontsize : float, default is None
        Currently, the fontsize is internally
        computed as a function of the dotsize,
        since the space available for the labels
        is determined by the dotsize. There's
        plans for user control in the future,
        but until then NotImplementedError will be thrown
    shortenAAs: boolean, default is True
        Use short AA-codes, e.g. E30 for GLU30. Only has effect if a topology
        is parsed
    aa_offset : int, default is 0
        Add this number to the resSeq value
    markersize : float, default is None
        The size of the dots. It is internally
        optimized to have adjacent dots fill
        the available space without overlapping
        among them. There's plans for user
        control in the future,  but until then
        NotImplementedError will be thrown
    bezier_linecolor : color-like, default is 'k'
        The color of the bezier curves connecting the residues.
        Can be a character, string or RGB value (not RGBA)
    plot_curves_only : bool, default is False
        Only plot the curves connecting the dots, but
        not the dots themselves or any other annotation.
        (labels, fragment names or SS information).
        The same caution as :obj:`iax` applies.
    textlabels : bool or array_like, default is True
        How to label the residue dots. Gets passed directly
        to :obj:`mdciao.flare.circle_plot_residues`.
        Options are:
         * True: the dots representing the residues
           will get a label automatically, either their
           serial index or the residue name, e.g. GLU30, if
           a :obj:`top` was passed.
         * False: no labeling
         * array_like : will be passed as :obj:`replacement_labels`
           to :obj:`mdciao.flare.add_fragmented_residue_labels`
    padding : iterable of len 3, default is [1,1,1]
        The padding, expressed as empty dot positions. Each number is
        used for:
         * the beginning of the flareplot, before the first residue
         * between fragments
         * at the end of the plot, after the last residue
    lw: float, default is None
        Line width of the contact lines
    signed_colors : dict, default is None
        Provide a color dictionary, e.g. {-1:"b", +1:"r"}
        to give different colors to positive and negative
        alpha values. If None, defaults to :obj:`bezier_linecolor`
    subplot : bool, default is False
        If True, the method checks if
        :obj:`iax` is the last axis in a
        figure (=all other panels have
        been already drawn) and then
        transfers the last plot's
        fontsizes and linewidths
        to panels (if possible).
        It will help produce more homogeneous
        plots when heuristics about font-sizing
        fail
    aura : iterable, default is None
        Scalar array, indexed with residue indices,
        e.g. RMSF, SASA, conv. degree...
        It will be drawn as an *aura* around the
        flareplot.
    coarse_grain : bool, default is False
        If True, will use the fragment definitions
        of :obj:`fragments` and/or `sparse_fragments`
        to coarse grain the frequencies into per-fragment
        frequencies and show them as a chord-diagram
        wrapping around :obj:`~mdciao.flare.freqs2chord`
        Check there for more info

    Returns
    -------
    iax : :obj:`~matplotlib.axes.Axes`
        You can do iax.figure.savefig("figure.png") to
        save the figure. Checkout
        :obj:`~matplotlib.figure.Figure.savefig` for more options

    plotted_pairs : 2D np.ndarray

    plot_attribs : dict
        Objects of the plot if the user wants
        to manipulate them further or re-use
        some attributes:
        "bezier_lw", "bezier_curves",
        "fragment_labels", "dot_labels",
        "dots", "SS_labels"
    """

    if _np.ndim(freqs)==1:
        freqs = _np.reshape(freqs,(1, -1))
    elif _np.ndim(freqs)==2:
        pass
    else:
        raise ValueError("Input array has to of shape either "
                         "(m) or (n, m) where n : n frames, and m: n_contacts")

    assert freqs.shape[1] == len(res_idxs_pairs), \
        "The size of the contact array and the " \
        "res_idxs_pairs array do not match %u vs %u"%(freqs.shape[1], len(res_idxs_pairs))

    # Figure out the combination of sparse_residues/fragments options
    residues_as_fragments, anchor_fragments, mute_fragments = \
        _futils._parse_residue_and_fragments(res_idxs_pairs,
                                             sparse_residues=sparse_residues,
                                             sparse_fragments=sparse_fragments,
                                             fragments=fragments,
                                             top=top,
                                             anchor_fragments=anchor_fragments,
                                             mute_fragments=mute_fragments)
    # Delete fragment names that won't be used
    if fragment_names is not None and fragments is not None:
        fragment_names = [fn for fr, fn in zip(fragments,fragment_names) if len(_np.intersect1d(fr,_np.hstack(residues_as_fragments)))>0]

    # Create a map
    residx2markeridx = _futils.value2position_map(_np.hstack(residues_as_fragments))

    if coarse_grain:
        assert len(residues_as_fragments)>1, "The fragment definitions of 'fragments' and/or 'sparse_fragments' can't be " \
                                             "used to coarse-grain the 'freqs'. Please check your input"
        fragment_colors = _futils.col_list_from_input_and_fragments(colors, residues_as_fragments, alpha=.80)
        fragment_colors = [fragment_colors[residx2markeridx[ff[0]]] for ff in residues_as_fragments]
        return freqs2chord(_np.average(freqs, axis=0), res_idxs_pairs, residues_as_fragments,
                           fragment_names=fragment_names,fragment_colors=fragment_colors,
                           panelsize=panelsize
                           )

    if plot_curves_only:
        plot_attribs = {}
        assert iax is not None, ("You cannot use "
                                 "plot_curves_only=True and iax=None. Makes no sense")
        xy = _futils.cartify_fragments(residues_as_fragments,
                                       r=r,
                                       angle_offset=angle_offset,
                                       padding=padding,
                                       )
        xy += center
    else:
        iax, xy, plot_attribs = circle_plot_residues(residues_as_fragments,
                                                     fontsize=fontsize,
                                                     colors=colors,
                                                     panelsize=panelsize,
                                                     padding=padding,
                                                     center=center,
                                                     ss_array=_get_SS(SS, top=top)[1],
                                                     fragment_names=fragment_names,
                                                     iax=iax,
                                                     markersize=markersize,
                                                     textlabels=textlabels,
                                                     shortenAAs=shortenAAs,
                                                     highlight_residxs=highlight_residxs,
                                                     aa_offset=aa_offset,
                                                     top=top,
                                                     r=r,
                                                     angle_offset=angle_offset,
                                                     aura=aura)

        circle_radius_in_pts = plot_attribs["dots"][0].radius * _points2dataunits(iax).mean()
        lw = circle_radius_in_pts # ??

    # All formed contacts
    ctcs_averaged = _np.average(freqs, axis=0)
    idxs_of_formed_contacts = _np.argwhere(_np.abs(ctcs_averaged)>freq_cutoff).squeeze()
    plot_this_pair_lambda = _futils.should_this_residue_pair_get_a_curve(residues_as_fragments,
                                                                         select_residxs=select_residxs,
                                                                         mute_fragments=mute_fragments,
                                                                         anchor_fragments=anchor_fragments,
                                                                         top=top, exclude_neighbors=exclude_neighbors)

    idxs_of_pairs2plot = _np.intersect1d(idxs_of_formed_contacts,
                                         [ii for ii, pair in enumerate(res_idxs_pairs) if plot_this_pair_lambda(pair)])

    if len(idxs_of_pairs2plot) > 0:
        pairs_of_nodes = [(xy[residx2markeridx[ii]],
                           xy[residx2markeridx[jj]]) for (ii, jj) in _np.array(res_idxs_pairs)[idxs_of_pairs2plot]]
        alphas = ctcs_averaged[idxs_of_pairs2plot]
        bezier_curves = add_bezier_curves(iax,
                                          pairs_of_nodes,
                                          alphas=alphas,
                                          center=center,
                                          lw=lw,
                                          bezier_linecolor=bezier_linecolor,
                                          signed_alphas=signed_colors
                                          )
        plot_attribs["bezier_lw"] = lw
        plot_attribs["bezier_curves"] = bezier_curves

    if subplot:
        if iax is iax.figure.axes[-1]:
            [_futils.fontsize_apply(iax, jax) for jax in iax.figure.axes]
            minwidth = min([_np.unique([line.get_linewidth() for line in jax.findobj(_Line2D)]) for jax in iax.figure.axes])
            [[line.set_linewidth(minwidth) for line in jax.lines] for jax in iax.figure.axes]
    return iax, idxs_of_pairs2plot, plot_attribs

def circle_plot_residues(fragments,
                         fontsize=None,
                         colors=True,
                         markersize=None,
                         r=1,
                         panelsize=4,
                         angle_offset=0,
                         padding=[1,1,1],
                         center=[0,0],
                         ss_array=None,
                         fragment_names=None,
                         iax=None,
                         textlabels=True,
                         shortenAAs=True,
                         highlight_residxs=None,
                         aa_offset=0,
                         top=None,
                         aura=None):
    r"""
    Circular background that serves as background for flare-plots. Is independent of
    the curves that will later be plotted onto it.

    Parameters
    ----------
    fragments : list
        List of iterables of residue idxs defining how the
        residues are split into fragments. If no
        :obj:`textlabels` are provided, the idxs
        themselves become the labels
    fontsize : float, default is None
        Currently, the fontsize is internally
        computed as a function of the dotsize,
        since the space available for the labels
        is determined by the dotsize. There's
        plans for user control in the future,
        but until then NotImplementedError will be thrown
    colors : can be of different types
        * False or None
            All returned colors will be the default color
        * True
            All returned colors will differ by fragment
        * string (anything matplotlib can understand as color)
            Use this color for all residues
        * iterable (array or list, not dict)
            len(colors) has to be either len(fragments) or
            sum([len(frag) for frag in fragments)
            In the first case, it's expanded to assign
            each residue in fragment the same color
            In the second case, since each residue in
            the fragments (apparently) has already a color,
            nothing happens
        * iterable (dict)
            Has to be of len(residxs_as_fragments)
    markersize : float, default is None
        The size of the dots. It is internally
        optimized to have adjacent dots fill
        the available space without overlapping
        among them. There's plans for user
        control in the future,  but until then
        NotImplementedError will be thrown
    r : scalar
        The radius of the circle, in axis inuts
    panelsize : float, default is 4
        The panelsize, in inches. Only has effect
        if :obj:`iax` is None and a new figure
        is created
    angle_offset : scalar
        Where the circle starts, in degrees. 0 means 3 o'clock,
        90 12 o'clock etc. It's the phi of polar coordinates
    padding : list, default is [1,1,1]
        * first integer : Put this many empty positions before the first dot
        * second integer: Put this many empty positions between fragments
        * third integer : Put this many empty positions after the last dot
    center : pair of floats
        where the circle is centered, in axis units
    ss_array : dict, list or array
        One-letter codes (H,B,E,C) denoting secondary
        structure. Has to be indexable by whatever
        indices are on :obj:`fragments`
    fragment_names : list
        The names of the fragments
    iax : :obj:`~matplotlib.axes.Axes`, default is None
        An axis to draw the dots on. It's parent
        figure has to have a tight_layout=True
        attribute. If no axis is passed,
        one will be created.
    textlabels : bool or array_like, default is True
        How to label the residue dots
         * True: the dots representing the residues
           will get a label automatically, either their
           serial index or the residue name, e.g. GLU30, if
           a :obj:`top` was passed.
         * False: no labeling
         * array_like : will be passed as :obj:`replacement_labels`
           to :obj:`mdciao.flare.add_fragmented_residue_labels`.
           These labels act as replacement and can cover all or just
           some residue, e.g. a mutated residue that you want
           to show as R38A instead of just A38, or use
           e.g. GPCR or CGN consensus labels.
    shortenAAs : boolean, default is True
        If :obj:`top` is not None, use "E50" rather than "GLU50"
    highlight_residxs : iterable of ints, default is None
        In case you don't want to construct a whole
        color list for :obj:`colors`, you can simply
        input a subset of :obj:`res_idxs` here and
        they will be shown in red.
    aa_offset : int, default is 0
        Add this number to the resSeq value
    top : :obj:`mdtraj.Topology`, default is None
        If provided, residue labels wil be auto-generated from
        here
    aura : iterable, default is None
        Scalar array, indexed with residue indices,
        e.g. RMSF, SASA, conv. degree...
        It will be drawn as an *aura* around the
        flareplot.

    Returns
    -------
    iax, xy, outdict

    outdict: dict
         Contains :obj:`matplotlib` objects
         like the dots and their labels:
         "fragment_labels", "dot_labels",
         "dots", "SS_labels", "r",
    """
    debug = False

    xy = _futils.cartify_fragments(fragments,
                                   r=r,
                                   angle_offset=angle_offset,
                                   padding=padding)
    xy += center

    n_positions = len(xy) + padding[0] + len(fragments)*padding[1] + padding[2]

    residues_to_plot_as_dots = _np.hstack(fragments)
    col_list = _futils.col_list_from_input_and_fragments(colors, fragments, alpha=.80)
    if iax is None:
        _plt.figure(figsize=(panelsize, panelsize), tight_layout=True)
        iax = _plt.gca()
    else:
        if not iax.figure.get_tight_layout():
            print("The passed figure was not instantiated with tight_layout=True\n"
                  "This may lead to some errors in the flareplot fontsizes.")
    # Do this first to have an idea of the points per axis unit necessary for the plot
    iax.set_xlim([center[0] - 3*r, center[0] + 3*r])
    iax.set_ylim([center[1] - 3*r, center[1] + 3*r])
    iax.set_aspect('equal')
    iax.draw(iax.figure.canvas.get_renderer())

    # Plot!
    if markersize is None:
        dot_radius = r * _np.sin(_np.pi/n_positions)
        dot_radius_in_pts = dot_radius*_points2dataunits(iax).mean()
        if dot_radius_in_pts < 1.5:
            print(ValueError("Drawing this many dots (%u residues + %u padding spaces) in a panel %3.1f inches wide/high "
                             "\nforces too small dotsizes and fontsizes. If crowding effects occur, either reduce the"
                             "\nnumber of residues or increase the panel size"%(len(xy), n_positions-len(xy), panelsize)))
        #TODO replace this with a call to RegularPolyCollection
        CPs = [_CP(ixy,
                   radius=dot_radius,
                   facecolor=col_list[ii],
                   edgecolor=None,
                   zorder=10) for ii, ixy in enumerate(xy)]
        [iax.add_artist(iCP) for iCP in CPs]
        outer_r_in_data_units = r + dot_radius
    else:
        raise NotImplementedError

    if debug:
        iax.add_artist(_plt.Circle(center,
                                   radius=outer_r_in_data_units,
                                   ec='r',
                                   fc=None,
                                   fill=False,
                                   zorder=10,
                                   lw=dot_radius_in_pts/10))
    # After the dots have been plotted,
    # we use their radius in points
    # as approximate fontsize
    if fontsize is None:
        fontsize = dot_radius_in_pts *1.75 # slightly smaller than the diameter
    else:
        raise NotImplementedError

    labels = []
    if textlabels:
        if isinstance(textlabels,bool):
            replacement_labels = None
        else:
            # Interpret the textlabels as replacements
            replacement_labels = textlabels
        outer_r_in_data_units += {True: 2,
                                  False:1}[dot_radius<1.5]*dot_radius
        labels = add_fragmented_residue_labels(fragments,
                                               iax,
                                               fontsize,
                                               center=center,
                                               r=outer_r_in_data_units,
                                               angle_offset=angle_offset,
                                               padding=padding,
                                               highlight_residxs=highlight_residxs,
                                               top=top,
                                               aa_offset=aa_offset,
                                               replacement_labels=replacement_labels)
        if debug:
            _futils._plot_fancypatches(labels, lw=dot_radius_in_pts / 10)
            iax.add_artist(_plt.Circle(center,
                                       radius=outer_r_in_data_units,
                                       ec='green',
                                       fc=None,
                                       fill=False,
                                       zorder=10,
                                       lw=dot_radius_in_pts / 10))

        delta_r = _futils._outermost_corner_of_fancypatches(labels)-outer_r_in_data_units
        outer_r_in_data_units += delta_r*.90 # fudge
        if debug:
            iax.add_artist(_plt.Circle(center,
                                       radius=outer_r_in_data_units,
                                       ec='b',
                                       fc=None,
                                       fill=False,
                                       zorder=10,
                                       lw=dot_radius_in_pts / 10))
    # Do we have SS dictionaries
    ss_labels = []
    if ss_array is not None:
        ss_colors = [_futils._SS2vmdcol[ss_array[res_idx]] for res_idx in residues_to_plot_as_dots]
        replacement_labels = {ii: ss_array[ii].replace("NA", " ").replace("E", "B") for ii in
                                 residues_to_plot_as_dots}
        ss_labels = add_fragmented_residue_labels(fragments,
                                                  iax,
                                                  fontsize * 2,
                                                  center=center,
                                                  r=outer_r_in_data_units,
                                                  angle_offset=angle_offset,
                                                  padding=padding,
                                                  shortenAAs=shortenAAs,
                                                  highlight_residxs=highlight_residxs, top=top,
                                                  aa_offset=aa_offset,
                                                  replacement_labels=replacement_labels,
                                                  colors=[_futils._make_color_transparent(col, .8, "w") for col in
                                                          ss_colors],
                                                  weight="bold")

        outer_r_in_data_units = _futils._outermost_corner_of_fancypatches(ss_labels)
        if debug:
            _futils._plot_fancypatches(ss_labels, lw=dot_radius_in_pts / 10)
            iax.add_artist(_plt.Circle(center,
                                       radius=outer_r_in_data_units,
                                       ec='purple',
                                       fc=None,
                                       fill=False,
                                       zorder=10,
                                       lw=dot_radius_in_pts / 10))

    if aura is not None:
        if debug:
            iax.add_artist(_plt.Circle([0, 0], outer_r_in_data_units,
                                       color=None,
                                       fill=False,
                                       ec="cyan",
                                       zorder=10,
                                       lw=dot_radius_in_pts / 10))
            iax.add_artist(_plt.Circle([0,0],outer_r_in_data_units, ec="k", alpha=.25, zorder=-100))

        auras, outer_r_in_data_units = _futils.add_aura(xy, aura[_np.hstack(fragments)], iax, outer_r_in_data_units+dot_radius,
                         [len(fr) for fr in fragments],
                         subtract_baseline=False
                         )

        if aura is not None:
            if debug:
                iax.add_artist(_plt.Circle([0, 0], outer_r_in_data_units,
                                           ec="g",
                                           facecolor=None,
                                           fill=False,
                                           zorder=10,
                                           lw=dot_radius_in_pts / 10))

    # Do we have names?
    frag_labels = []
    if fragment_names is not None:
        span = 2 * outer_r_in_data_units
        frag_fontsize_in_aus =  span/6 * 1/5 # (average_word_length, fraction of panel space)
        frag_fontsize_in_pts = _np.max((3*fontsize, frag_fontsize_in_aus * _points2dataunits(iax).mean()))
        outer_r_in_data_units += frag_fontsize_in_aus
        if debug:
            iax.add_artist(_plt.Circle(center,
                                       radius=outer_r_in_data_units,
                                       ec='pink',
                                       fc=None,
                                       fill=False,
                                       zorder=10,
                                       lw=dot_radius_in_pts / 10))
        frag_labels = _futils.add_fragment_labels(fragments,
                                                  [replace4latex(str(ifrag)) for ifrag in fragment_names],
                                                  iax,
                                                  angle_offset=angle_offset,
                                                  padding=padding,
                                                  fontsize=frag_fontsize_in_pts,
                                                  center=center,
                                                  r=outer_r_in_data_units
                                                  )

        outer_r_in_data_units = _futils._outermost_corner_of_fancypatches(frag_labels)

        if debug:
            iax.add_artist(_plt.Circle(center,
                                       radius=outer_r_in_data_units,
                                       ec='orange',
                                       fc=None,
                                       fill=False,
                                       zorder=10,
                                       lw=dot_radius_in_pts / 10))

    iax.set_yticks([])
    iax.set_xticks([])
    _futils.change_axlims_and_resize_Texts(iax, outer_r_in_data_units)
    _futils.un_overlap_via_fontsize(frag_labels, fac=.90)

    return iax, xy, {"fragment_labels":frag_labels,
                     "dot_labels":labels,
                     "dots":CPs,
                     "SS_labels":ss_labels,
                     "r": outer_r_in_data_units
                     }


def add_bezier_curves(iax,
                      nodepairs_xy,
                      alphas=None,
                      center=[0,0],
                      lw=1,
                      bezier_linecolor='k',
                      signed_alphas=None,
                      correct_adjacent_nodes=True
                      ):
    r"""
    Generate and plot bezier curves using :obj:`bezier.Curves` as a base class

    Parameters
    ----------
    iax : :obj:`~matplotlib.axes.Axes`
        The axes to draw on
    nodepairs_xy : iterable of pairs of pairs of floats
        Each item is a pair of pairs [(x1,y1),(x2,y2)]
        to be connected with bezier curves 1<--->2
    alphas : iterable of floats, default is None
        The opacity of the curves connecting
        the pairs in :obj:`nodepairs_xy`.
        If None is provided, it will be set to 1
        for all curves. If provided,
        must be of len(nodepairs_xy).
    center : array-like with two floats, default is [0,0]
        The center of the flareplot, in
        data units
    lw : int, default is 1
        Linewidth of the curves, in pts
    bezier_linecolor: :obj:`matplotlib` color, default is "k"
        The  color of the curves
    signed_alphas : dict, default is None
        Provide a color dictionary, e.g. {-1:"b", +1:"r}
        to give different colors to positive and negative
        alpha values. Overwrites whatever is in :obj:`bezier_linecolor`.
        If None, defaults to :obj:`bezier_linecolor`
    correct_adjacent_nodes : boolean, default is True
        If two nodes are too close to each other,
        use a shifted center for the Bezier curve so
        that it's visually easier to find. Currently,
        adjacent is hard-coded to mean "nodes are ten
        times closer to each other than to the center"

    Returns
    -------
    bz_curves : list
        The :obj:`bezier.Curves` plotted

    """
    if alphas is None:
        alphas = _np.ones(len(nodepairs_xy)) * 1.0

    assert len(alphas)==len(nodepairs_xy)

    if signed_alphas in [None,{}]:
        signed_alphas={-1:bezier_linecolor,
                       +1:bezier_linecolor}

    bz_curves=[]
    for nodes, ialpha in zip(nodepairs_xy, alphas):

        d_nodes = _np.sqrt(((nodes[0] - nodes[1]) ** 2).sum())
        r = _np.mean([_np.linalg.norm(nodes[0] - center), _np.linalg.norm(nodes[1] - center)])
        pair_center = (nodes[0] + nodes[1]) / 2
        pc2c = center - pair_center
        if correct_adjacent_nodes and r / 10 > d_nodes:
            _center = pair_center + pc2c / 10
        else:
            _center = center
        bz_curves.append(_futils.create_flare_bezier_2(nodes, center=_center))
        bz_curves[-1].plot(50,
                           ax=iax,
                           alpha=_np.abs(ialpha),
                           color=signed_alphas[_np.sign(ialpha)],
                           lw=lw,  # _np.sqrt(markersize),
                           zorder=-1
                           )

    return bz_curves

def add_fragmented_residue_labels(fragments,
                                  iax,
                                  fontsize,
                                  center=[0,0],
                                  r=1,
                                  angle_offset=0,
                                  padding=[0,0,0],
                                  **add_res_labs_kwargs,
                                  ):
    r"""
    Like add_labels but starting from the fragments

    Parameters
    ----------
    fragments : iterable of iterables
        List of iterables of residue idxs defining how the
        residues are split into fragments. If no
        :obj:`textlabels` are provided, the idxs
        themselves become the labels
    iax : :obj:`~matplotlib.axes.Axes`
    fontsize : int
    r : scalar
        The radius of the circle, in axis inuts
    angle_offset : scalar
        Where the circle starts, in degrees. 0 means 3 o'clock,
        90 12 o'clock etc. It's the phi of polar coordinates
    padding : list, default is [1,1,1]
        * first integer : Put this many empty positions before the first dot
        * second integer: Put this many empty positions between fragments
        * third integer : Put this many empty positions after the last dot
    center : pair of floats
        where the circle is centered, in axis units
    add_res_labs_kwargs

    Returns
    -------
    labels : list of text objects

    """
    xy_labels, xy_angles = _futils.cartify_fragments(fragments,
                                                     r=r,
                                                     return_angles=True,
                                                     angle_offset=angle_offset,
                                                     padding=padding)
    xy_labels += center

    residues_to_plot_as_dots = _np.hstack(fragments)
    labels = _futils.add_residue_labels(iax,
                                        xy_labels,
                                        residues_to_plot_as_dots,
                                        fontsize,
                                        center=center,
                                        **add_res_labs_kwargs,
                                        )

    return labels

def freqs2chord(freqs, res_idxs_pairs, fragments,
                fragment_names=None,
                iax=None,
                panelsize=10,
                fragment_colors=None,
                add_sigma=True,
                normalize_to_sigma=None,
                min_sigma=0,
                clockwise=True,
                ):
    r"""
    Coarse-grain the per-residue frequencies into a per-fragment `chord-diagram <https://en.wikipedia.org/wiki/Chord_diagram>`_.

    Can be thought of as a coarse-grained flareplot, where
    residues belonging to the same fragments get turned into
    arcs instead of being plotted as individual dots.
    Each arc-length is proportional to the fragments' overall number of
    contacts, NOT to the number of residues that the arc contains.

    If fragments don't participate in any contacts, they
    are hidden, i.e. the plot is always 'sparse'

    Curves connecting residue-pairs are squashed into chords
    connecting the arc. The opacity of each chord does
    NOT represent the underlying frequencies.

    The main difference with :obj:`mdciao.flare.freqs2flare`
    is that the per-fragment number of contacts gets represented
    visually through the length the arc and that the chords
    don't encode the number of underlying contacts.

    This method uses the `mpl_chord_diagram <https://codeberg.org/tfardet/mpl_chord_diagram>`_
    package by `@tfardet <https://codeberg.org/tfardet>`_
    with contributions from `Guillermo Perez-Hernandez <https://codeberg.org/gph82>`_.

    Parameters
    ----------
    freqs : iterable of floats
        A list of floats representing frequencies
    res_idxs_pairs : list of pairs of integers, default is None
        The pairs of residue indices behind each freq entry
        in `freqs`. They can be any integers that could represent
        a residue. The only hard condition is that the set
        of np.unique(res_idxs_pairs) must be contained
        within `fragments`
    fragments : a list of lists of integers
        These are the fragments used to coarse-grain
        the `freqs`. Any residue index of `res_idxs_pairs`
        that has a non-zero frequency associated to it
        in `freqs` has to appear in one (and only one)
        of the `fragments`.
    fragment_names : list, default is None
        The fragment names
    iax: :obj:`~matplotlib.axes.Axes`, default is None
        Parse an axis to draw on, otherwise one will be created
        using `panelsize`.
    normalize_to_sigma : float, default is None
        If a float is provided, then the arcs
        of the chord-diagram can occupy the
        full 360% degrees of the circumference if,
        and only if, the sum over all per-fragment
        contacts (Sigma) aggregates to `normalize_sigma`.
        If they aggregate to a smaller value, then
        they will occupy a proportionately smaller
        portion of the circumference. If they aggregate
        to a larger value, then a error will be thrown.
        The default behavior (None) is to always occupy
        the full 360°, but this makes it hard to
        compare across different chord-diagrams for systems
        with different Sigmas, since they are scaled
        to 360° by default.
    min_sigma : float, default is 0
        `fragments` with less contacts than this, i.e.
        aggregating to Sigma smaller than this,
        aren't plotted. Please note that this
        cutoff applies only to the aggregated, per-fragment value, i.e.
        if enough individual contacts, each with frequencies
        < min_sigma, appear in one fragment and aggregate
        to a value > min_sigma, then this fragment is plotted.
        fragments dropped because of `min_sigma`, aren't
        considered when computing how much of the circumference
        is occupied by the diagram, i.e.  the Sigma used
        to `normalize_to_sigma` doesn't include them.
    panelsize : int, default is 10
        The panelsize for the plot, in inches
    clockwise : bool, default is True
        Plot fragments (=arcs)
        clock-wise from 0° (3 o'clock),
        the same way that the flareplots do.
        If False, progress counter-clock-wise.
        The order in which the fragments are
        plotted is the order in which they
        appear in `fragments`, regardless of
        what indices they contain.
    fragment_colors : iterable of color-likes
        The colors given to the arcs representing
        the fragments
    add_sigma : bool, default is True
        Add the sum of freqs per fragment
        to the arc

    Returns
    -------
    iax : :obj:`~matplotlib.axes.Axes`
    non_zeros : list
        The idxs of :obj:`fragments`
        that have been plotted
    plot_attribs : dict
        Objects of the plot, for the user to
        to manipulate further. Some attributes:
        "sigmas", "fragment_names", "fragment_labels",
        "dots", "r". The attribute "dots" is a shim namedtuple
        with one single attribute, "radius",
        for compatibility with downstream
        manipulations.
    """
    # Coarse grain the matrix
    mat_CG = _futils.coarse_grain_freqs_by_frag(freqs, res_idxs_pairs, fragments)
    # Get the sparse matrix (element-wise)
    sparse_mat, non_zeros = _futils.sparsify_sym_matrix(mat_CG)

    non_zero_fragment_names = [[fragment_names[ii] for ii in non_zeros] if fragment_names is not None else None][0]
    non_zero_fragment_colors = [[fragment_colors[ii] for ii in non_zeros] if fragment_colors is not None else None][0]

    # Get the sparse matrix (row/column wise)
    sparse_mat, non_zeros, discarded = _futils.sparsify_sym_matrix_by_row_sum(sparse_mat, min_sigma)

    non_zero_fragment_names = \
        [[non_zero_fragment_names[ii] for ii in non_zeros] if fragment_names is not None else None][0]
    non_zero_fragment_colors = \
        [[non_zero_fragment_colors[ii] for ii in non_zeros] if fragment_colors is not None else None][0]

    order = _np.arange(len(non_zeros))
    if clockwise:
        order = order[::-1]

    if normalize_to_sigma:
        assert _np.round(normalize_to_sigma, 6) >= _np.round(sparse_mat.sum() / 2, 6), \
            ValueError("The Sigma value for normalizing has to be larger than "
                       "the aggregated `freqs`, but `normalize_to_sigma` = %4.2f < %4.2f = sum(freqs) "%
                       (_np.round(normalize_to_sigma, 6), _np.round(sparse_mat.sum() / 2, 6)))
        extent = 360 * sparse_mat.sum() / 2 / normalize_to_sigma
    else:
        extent = 360

    # Grab text objects already in the input axis
    old_texts = []
    if iax is None:
        _plt.figure(figsize=(panelsize, panelsize), tight_layout=True)
        iax = _plt.gca()
        old_texts = list(iax.texts)

    res = _chord_diagram(sparse_mat,
                         chord_colors="gray",
                         order=order,
                         names=non_zero_fragment_names,
                         colors=non_zero_fragment_colors,
                         ax=iax,
                         extent=extent)

    # Grab the new text-labels and un-overlap them
    frag_labels = [txt for txt in iax.texts if txt not in old_texts]  # only grab new text-objects
    [t.set_bbox({"boxstyle": "square,pad=0.5", "fc": "blue", "ec": "none", "alpha": .0}) for t in frag_labels]
    _futils.un_overlap_via_fontsize(frag_labels, fac=.85)

    if add_sigma:
        sigma_values = ["$\Sigma$ = %3.1f" % (ifreq) for ifreq in sparse_mat.sum(0)]
        sigma_labels = []
        for ii, rr in enumerate(res):
            __, psi = _futils.cart2pol(rr[0],rr[1])
            x, y = _futils.pol2cart(1 - .1 / 2, psi)
            sigma_labels.append(iax.text(x, y, sigma_values[order[ii]],
                     rotation_mode="anchor", ha="center", va="center", rotation=rr[-1],
                                         bbox={"boxstyle": "square,pad=0.0", "fc": "white", "ec": "none", "alpha": .5}))
        _futils.un_overlap_via_fontsize(sigma_labels, fac=.90)

    # Create fake "dot" object with a  .radius attribute for compatibility with the downstream _utils.add_parent_labels
    dot = _namedtuple("fake_dot", ["radius"])
    return_dict = {"fragment_names": non_zero_fragment_names,
                   "fragment_labels": frag_labels,
                   "r": _futils._outermost_corner_of_fancypatches(frag_labels),
                   "dots": [dot(_signature(_chord_diagram).parameters["width"].default / 4)],
                   "sigmas": sparse_mat.sum(0)} # we don't need [order[::-1]] here!!
    if add_sigma:
        return_dict["sigma_labels"] = sigma_labels

    return iax, non_zeros, return_dict



#!/usr/bin/env python3

##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#    
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from mdciao.examples.examples import ExamplesCLTs
from mdciao.parsers import parser_for_examples, _inform_of_parser
from os import path
from fnmatch import filter as _fnfilter
execute = True
list = False
import sys
parser = parser_for_examples()
if len(sys.argv)==1:
    args = parser.parse_args(args=["-h"])
else:
    args = parser.parse_args(args=None)

#print(_inform_of_parser(parser))
#args = parser.parse_args()

clt = args.clt
ex = ExamplesCLTs(short=args.short)

if args.clt.endswith(".py"):
    clt = path.splitext(args.clt)[0]
clts = _fnfilter(ex.clts,"*%s*"%clt)
if len(clts)==0:
    print("'%s' did not return any command-line-tool."%clt)
    print(parser.epilog)
for clt in clts:
    ex.show(clt)
    if args.execute:
        ex.run(clt, show=False)



r"""
Simple operations with bonds between the
residues of an :obj:`mdtraj.Topology`

.. autosummary::
   :nosignatures:
   :toctree: generated/


"""
import numpy as _np
from scipy.sparse.csgraph import connected_components as _concom

def connected_sets(mat):
    r"""
    Return the connected components/sets of a an adjacency matrix

    Uses :obj:`~scipy.sparse.csgraph.connected_components`
    under the hood with directed=False

    Parameters
    ----------
    mat : 2D _np.array, square matrix (M,M)
        Adjacency matrix, can be symmetric or not.
        Nodes are always self-adjacent, i.e.
        the diagonal of :obj:`mat` is ignored

    Returns
    -------
    sets: list
        The connected components as 1D _np.ndarrays
    """
    sets = []
    nsets, labels = _concom(mat, directed=False)
    for ii in range(nsets):
        sets.append(_np.flatnonzero(labels==ii))
    return sets

def top2residue_bond_matrix(top,
                            force_resSeq_breaks=False,
                            verbose=True,
                            create_standard_bonds=False):
    r"""Return a symmetric residue-residue bond matrix from a :obj:`~mdtraj.Topology`.

    The bonds used are those found in :obj:`~mdtraj.Topology.bonds`

    Parameters
    ----------
    top : :obj:`mdtraj.Topology`
    force_resSeq_breaks : boolean, default is False
        Delete bonds if there is a resSeq jump between residues.
    verbose : boolean, default is True
        Print a statement if residue index has no bonds
    create_standard_bonds : boolean, default is False
        Advanced users only, can easily lead to wrong
        results in case of .gro files, because
        :obj:`~mdtraj.Topology.create_standard_bonds`
        needs chain information to avoid creating
        bonds between residues that follow one another
    Returns
    -------
    residue_bond_matrix : 2D np.ndarray
        Returns a symmetric adjacency matrix with entries ij=1 and ji=1,
        if there is a bond between residue i and residue j.
    """

    if len(top._bonds) == 0:
        if create_standard_bonds:
            top.create_standard_bonds()
        else:
            raise ValueError("\nThe parsed topology does not contain any bonds in the top._bonds attribute.\n"
                             " * If you are using low-level methods, like mdciao.utils.bonds.top2residue_bond_matrix,\n"
                             "   try using 'create_standard_bonds=True'. This will 'create bonds based on the atom\n"
                             "   and residue names for all standard residue types'.\n"
                             " * If you are using CLI methods like mdciao.cli.residue_neighborhoods or mdc_neighborhoods.py,\n"
                             "   try using 'naive_bonds=True'. This will create bonds between adjacent residues regardless of\n"
                             "   atom-types.\n"
                             "Please read the docs of the different options for potential pitfalls and risks!")

    residue_bond_matrix = _np.zeros((top.n_residues, top.n_residues), dtype=int)
    for ibond in top._bonds:
        r1, r2 = ibond.atom1.residue.index, ibond.atom2.residue.index
        rSeq1, rSeq2 = ibond.atom1.residue.resSeq, ibond.atom2.residue.resSeq
        residue_bond_matrix[r1, r2] = 1
        residue_bond_matrix[r2, r1] = 1
        if force_resSeq_breaks and _np.abs(rSeq1 - rSeq2) > 1:  # mdtrajs bond-making routine does not check for resSeq
            residue_bond_matrix[r1, r2] = 0
            residue_bond_matrix[r2, r1] = 0
    for ii, row in enumerate(residue_bond_matrix):
        if row.sum()==0 and verbose:
            print("Residue with index %u (%s) has no bonds to other residues"%(ii,top.residue(ii)))

    return residue_bond_matrix

#TODO do we need this anymore?
def top2residuebonds(top,**top2residue_bond_matrix_kwargs):
    return _residue_bond_matrix_to_triu_bonds(top2residue_bond_matrix(top, **top2residue_bond_matrix_kwargs))

#TODO do we need this anymore?
def _residue_bond_matrix_to_triu_bonds(residue_bond_matrix):
    _np.testing.assert_array_equal(residue_bond_matrix, residue_bond_matrix.T), \
        ("This is not a symmetric residue bond matrix\n", residue_bond_matrix)
    bonds = []
    for ii, jj in _np.vstack(_np.triu_indices_from(residue_bond_matrix, k=1)).T:
        if residue_bond_matrix[ii, jj] == 1:
            bonds.append([ii, jj])
    return bonds

def bonded_neighborlist_from_top(top, n=1, verbose=False):
    """
    Bonded neighbors of all the residues in the topology file.

    Parameters
    ----------
    top : :obj:`~mdtraj.Topology`
    n : int, default is 1
        Number of bonded neighbors considered, i.e.
        A-B-C has only [B] as neighbors if :obj:`n` = 1,
        but [B,C] if :obj:`n` = 2

    Returns
    -------
    neighbor_list : list of lists
        Lisf of len top.n_residues. The i-th  list
        contains the :obj:`n` -bonded neighbors of
        the i-th residue, in ascending order
    """

    residue_bond_matrix = top2residue_bond_matrix(top,verbose=verbose)
    return neighborlists_from_adjacency_matrix(residue_bond_matrix, n)

#todo this is very slow and a bit of overkill if one is only interested in the
#neighbors of one particular residue
# Do it graph-thetically, ideally
def neighborlists_from_adjacency_matrix(mat, n):
    r"""
    Return neighborlists from an adjacency matrix.

    The diagonal of :obj:`mat` is ignored, i.e. it can be 0 or 1

    Parameters
    ----------
    mat : 2D _np.array, square matrix (M,M)
        Adjacency matrix, can be symmetric or not
    n : int
        Connectedness.

    Returns
    -------
    neighbors : list
        A list of len M where the i-th entry
        contains the indices of the nodes
        separated from the i-ith node by
        a maximum of :obj:`n` jumps. The indices
        are always in ascending order
    """
    neighbor_list = [[ii] for ii in range(mat.shape[0])]
    for kk in range(n):
        for ridx, ilist in enumerate(neighbor_list):
            new_neighborlist = [ii for ii in ilist]
            for rn in ilist:
                row = mat[rn]
                bonded = _np.flatnonzero(row)
                toadd = [nn for nn in bonded if nn not in ilist and nn != ridx]
                if len(toadd):
                    # print("neighbor %u adds new neighbor %s:"%(rn, toadd))
                    new_neighborlist += toadd
                    # print("so that the new neighborlist is: %s"%new_neighborlist)

            neighbor_list[ridx] = [ii for ii in _np.unique(new_neighborlist) if ii != ridx]
            # break

    # Check that the neighborlist works both ways
    for ii, ilist in enumerate(neighbor_list):
        for nn in ilist:
            assert ii in neighbor_list[nn]

    return neighbor_list

def top2residue_bond_matrix_naive(top, only_protein=True, fragments=None):
    r""" Creates a naive (=linear) residue-residue bond-matrix,
    where a bond is assumed between the n-th and the n+1-th residue

    Usually, one would resort to this method if the topology's
    own create_standard_bonds() method does not work, e.g.
    because there's only alpha Carbons

    Parameters
    ----------
    top : :obj:`~mdtraj.Topology`
    only_protein : bool, default is True
        Only create bonds when both residues
        satisfy are protein residues. The
        check is done using the attribute
        :obj:`mdtraj.core.topology.Residue.is_protein`
    fragments : iterable of ints, default is None
        Use fragment/chain definition to avoid
        bonds between fragments/chains. These
        definitions need to cover the entire :obj:`~mdtraj.Topology`,
        i.e. each residue index must appear once (an and only once)
        in the :obj`fragments`. Note that breaks
        introduced via the :obj:`only_protein argument`
        will always be present, whether these residues
        are in the same fragment or not

    Returns
    -------
    mat : 2D np.ndarray
        Symmetric residue-residue bond-matrix.
        The diagonal is filled with ones as well.
    """
    bonds = _np.zeros((top.n_residues, top.n_residues), dtype=int)

    if fragments is None:
        fragments = [_np.arange(top.n_residues)]

    #TODO this fragment checking might defined elsewhere as a function
    # but I don't want to introduce any other dep here
    stacked = _np.concatenate(fragments)
    unique = _np.unique(stacked)
    missing = [ii for ii in range(top.n_residues) if ii not in stacked]
    if len(missing)>0:
        raise ValueError("Bad fragment definition. Residue idxs missing: %s"%missing)
    if len(unique)<len(stacked):
        raise ValueError("Bad fragment definition. Residue idxs repeated: %s"%_np.flatnonzero(_np.bincount(stacked)!=1))

    if only_protein is True:
        protein = [rr.index for rr in top.residues if rr.is_protein]
    else:
        protein = [rr.index for rr in top.residues]

    for ifrag in fragments:
        prot_frag = _np.intersect1d(ifrag, protein)
        for ii in prot_frag:
            if ii+1 in prot_frag:
                bonds[ii,ii+1] = 1
                bonds[ii+1,ii] = 1

    _np.fill_diagonal(bonds,1)
    return bonds



r"""

RCSB-PDB web lookups

Relevant links:
 * https://data.rcsb.org/index.html#data-api
 * https://data.rcsb.org/data-attributes.html

Methods not documented here are either private or will be soon,
please note the API is not yet stable.

.. currentmodule:: mdciao.pdb

.. autosummary::
   :toctree: generated/

    pdb2ref
    pdb2traj


"""
from .pdb import *


r"""
Programatic access to the CLI methods

.. currentmodule:: mdciao.cli

.. autosummary::
   :toctree: generated/

   residue_neighborhoods
   interface
   sites
   compare
   residue_selection
   pdb
   fragment_overview
"""
from .cli import *



##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import numpy as _np

from ._textutils import \
    outermost_text_corner as _outermost_corner_of_fancypatches, \
    any_overlap_via_FancyBoxPach

from mdciao.plots.plots import _colorstring
from mdciao.utils.bonds import bonded_neighborlist_from_top
from mdciao.utils.lists import \
    assert_no_intersection as _no_intersect, \
    re_warp as _re_warp, \
    find_parent_list as _find_parent_list
from mdciao.fragments import assign_fragments
from mdciao.utils.str_and_dict import replace4latex as _replace4latex
from matplotlib.colors import to_rgb as _to_rgb
from matplotlib.collections import LineCollection as _LCol

from collections import Counter as _Counter, defaultdict as _defdict
from matplotlib.patches import CirclePolygon as _CP, Polygon as _PG

#https://stackoverflow.com/a/33375738
def _make_color_transparent(color_fg, alpha, color_bg="white"):
    r"""Return a color that's a three-value RGB equivalent to a four valued (rgb+alpha) transparent color"""
    return [alpha * c1 + (1 - alpha) * c2 for (c1, c2) in zip(_to_rgb(color_fg), _to_rgb(color_bg))]

_mycolors = _colorstring.split(",")

from bezier import Curve as _BZCurve
class my_BZCURVE(_BZCurve):
    """
    Modified Bezier curve to plot with line-width and to have self.Line2D
    """

    def plot(self, num_pts, color=None, alpha=None, ax=None,lw=1, zorder=None):
        """Plot the current curve.

        Args:
            num_pts (int): Number of points to plot.
            color (Optional[Tuple[float, float, float]]): Color as RGB profile.
            alpha (Optional[float]): The alpha channel for the color.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.

        Raises:
            NotImplementedError: If the curve's dimension is not ``2``.
        """
        from bezier import _plot_helpers
        if self._dimension != 2:
            raise NotImplementedError(
                "2D is the only supported dimension",
                "Current dimension",
                self._dimension,
            )

        s_vals = _np.linspace(0.0, 1.0, num_pts)
        points = self.evaluate_multi(s_vals)
        if ax is None:
            ax = _plot_helpers.new_axis()
        self._Line2D = ax.plot(points[0, :], points[1, :], color=color, alpha=alpha, lw=lw, zorder=zorder)[0]
        return ax

    @property
    def Line2D(self):
        return self._Line2D

def create_flare_bezier(nodes, center=None):
    middle = _np.floor(len(nodes) / 2).astype("int")
    if center is not None:
        nodes = _np.vstack((nodes[:middle], center, nodes[middle:]))
    return my_BZCURVE(nodes.T, degree=2)

def create_flare_bezier_2(nodes,
                        center):
    r"""

    Parameters
    ----------
    nodes : 2D np.ndarray
        The x-y positions of
        the nodes to be joined
        via a Bezier curve
    center : pair of floats
        x-y coordinate of the center

    Returns
    -------

    bzc : :obj:`my_BZCURVE`

    """
    assert len(nodes)==2
    nodes = _np.vstack((_np.array((nodes[0], center, nodes[1]),dtype="float")))
    return my_BZCURVE(nodes.T, degree=2)

def regspace_angles(npoints, circle=360, clockwise=True):
    r"""Return the angular values for spreading n points equidistantly in a circle

    Note
    ----
    The goal of this is to work together with :obj:`cartify_fragments`, s.t.
     * offset=0 puts the first "angle" at 0 s.t that the first cartesian
     pair will be (1,0) and then clockwise on (0,-1) etc (see :obj:`pol2cart`)
     * even if :obj:`circle`=360, the last entry of the returned :obj:`angles`
       will not be 360, regardless of the :obj:`offset`

    Parameters
    ----------
    n : integer
        number of points
    circle : float, default is 360
        size of the circle in degrees
    clockwise : boolean, default is True
        In which direction the angle array grows.

    Returns
    -------
    angles : np.ndarray of len(array_in)

    """
    npoints
    dang = circle / npoints
    if clockwise:
        fac=-1
    else:
        fac=1
    angles = _np.hstack([fac*(ii * dang) for ii in range(npoints)])
    return angles

def fragment_selection_parser(fragments,
                              hide_fragments=None,
                              only_show_fragments=None):
    r"""
    Return a lambda that will decide whether a residue pair should be shown or not based on a fragment selection.

    Note
    ----
    At least one of :obj:`hide_fragments` or :obj:`only_show_fragments` should
    be None, i.e. the user can either make no choice (==> all residue pairs
    should be plotted) or select a subset of residue pairs to show by
    either `hiding` some or `only_showing' some fragments, but not muting or showing
    simultaneously

    Parameters
    ----------
    fragments : iterable
        Fragment list, each item is itself an iterable of residue indices
    hide_fragments : iterable, default is None
        subset of :obj:`fragments` to mute, i.e. residues belonging
        to these fragments will be deleted from the output
    only_show_fragments : iterable, default is None
        show  subset of :obj:`fragments` to force-show, i.e. residues belonging
        to these fragments will be shown even if they are originally
        muted

    Returns
    -------
    condition : lambda : residx_pair -> bool
        Take a residx_pair and tell whether it should be plotted
        according to the fragment selection
    """

    if hide_fragments is None and only_show_fragments is None:
        condition = lambda res_pair : True
    elif hide_fragments is None and only_show_fragments is not None:
        residxs = _np.hstack([fragments[iseg] for iseg in only_show_fragments])
        condition = lambda res_pair: _np.any([ires in residxs for ires in res_pair])
    elif hide_fragments is not None and only_show_fragments is None:
        residxs = _np.hstack([fragments[iseg] for iseg in hide_fragments])
        condition = lambda res_pair: _np.all([ires not in residxs for ires in res_pair])
    else:
        raise ValueError("'hide_fragments' and 'only_show_fragments' can't simultaneously be 'None'")

    return condition

def cartify_fragments(fragments,
                      r=1.0,
                      return_angles=False,
                      angle_offset=0,
                      **padding_kwargs):
    r"""
    Cartesian coordinates on a circle of radius :obj:`r` for each sub-item in the lists of :obj:`fragments`

    Note
    ----
    works internally in radians

    Parameters
    ----------
    fragments: iterable of iterables
        They represent the fragments
    r : float, radius
    return_angles : bool, default is False
    angle_offset : float
        Degrees. Where in the circle the first sub-item in of
        the first item of :obj:`fragments` should start. Default
        is 0 which places the first point at xy = r,0 (3 o'clock)
    padding_kwargs : optional keyword args for :obj:`pad_fragment_positions`

    Returns
    -------

    """

    padded_fragment_list = pad_fragment_positions(fragments,**padding_kwargs)
    spaced_segments = _np.hstack(padded_fragment_list)
    spaced_angles = regspace_angles(len(spaced_segments),
                                    circle=2 * _np.pi,
                                    )+angle_offset * _np.pi / 180
    angles = _np.hstack([spaced_angles[ii] for ii, idx in enumerate(spaced_segments) if idx is not None])
    xy = pol2cart(_np.ones_like(angles) * r, angles)

    if not return_angles:
        return _np.vstack(xy).T
    else:
        return _np.vstack(xy).T, angles

def pad_fragment_positions(fragments,
                           padding=[0,0,0],
                           pad_value=None):
    r"""
    Provided a list of fragments, introduce a number of :obj:`pad_value` between them

    Parameters
    ----------
    fragments : list of iterables
    padding : list of ints, default is [0,0,0]
        Each integer controls the number of padding values used:
        * at the very beginning of the fragment list
        * at the beginning of each fragment
        * at the very end of the fragment list

    pad_value : arbritary, default is None
        What value to pad with (None, np.NaN, 0, "X") etc

    Returns
    -------
    padded_fragments : list
    """
    padding_initial = [pad_value for ii in range(padding[0])]
    padding_final = [pad_value for ii in range(padding[2])]
    padded_frags = [_np.hstack(([pad_value] * padding[1], iseg)) for iseg in fragments]

    if len(padding_initial) > 0:
        padded_frags = padding_initial + padded_frags
    if len(padding_final) > 0:
        padded_frags += padding_final

    return padded_frags

def pol2cart(rho, phi):
    r"""
    Polar to cartesian coordinates. Angles in radians

    Parameters
    ----------
    rho : float
    phi : float

    Returns
    -------
    x, y
    """
    x = rho * _np.cos(phi)
    y = rho * _np.sin(phi)

    return (x, y)


# todo CLEAN THIS COLOR MESS
def col_list_from_input_and_fragments(colors,
                                      residxs_as_fragments,
                                      default_color="gray",
                                      alpha=1,
                                      ):
    r"""
    per-residue color list taking possible fragmentation into account


    Parameters
    ----------
    colors : can be of different types
        * False or None
            All returned colors will be the default color
        * True
            All returned colors will differ by fragment
        * string (anything matplotlib can understand as color)
            Use this color for all residues
        * iterable (array or list, not dict)
            len(colors) has to be either len(fragments) or
            sum([len(frag) for frag in fragments)
            In the first case, it's expanded to assign
            each residue in fragment the same color
            In the second case, since each residue in
            the fragments (apparently) has already a color,
            nothing happens
        * iterable (dict)
            Has to be of len(residxs_as_fragments)
    res_idxs : iterable of ints or iterable thereof
        The residues for which to generate the
        colors. If an iterable of ints is passed,
        only one fragment is assumed

    Returns
    -------
    colors : list
        list of len(_np.hstack(fragments))

    """
    if colors is None:
        colors=False

    if isinstance(residxs_as_fragments[0], int):
        residxs_as_fragments = [residxs_as_fragments]

    if isinstance(colors, bool):
        if colors:
            to_tile = _mycolors
        else:
            to_tile = [default_color]
        jcolors = _np.tile(to_tile, _np.ceil(len(residxs_as_fragments) / len(to_tile)).astype("int"))
        col_list = _np.hstack([[jcolors[ii]] * len(iseg) for ii, iseg in enumerate(residxs_as_fragments)])
    elif isinstance(colors, str):
        to_tile = [colors]
        jcolors = _np.tile(to_tile, _np.ceil(len(residxs_as_fragments) / len(to_tile)).astype("int"))
        col_list = _np.hstack([[jcolors[ii]] * len(iseg) for ii, iseg in enumerate(residxs_as_fragments)])

    elif isinstance(colors, (list, _np.ndarray)):
        if len(colors) == len(residxs_as_fragments):
            col_list = _np.vstack([_np.vstack([colors[ii]]*len(fr)) for ii, fr in enumerate(residxs_as_fragments)]).squeeze().tolist()
        elif len(colors) == len(_np.hstack(residxs_as_fragments)):
            col_list = colors
        else:
            raise ValueError("The number of input colors (%u) "
                             "doesn't match the number of fragments (%u) "
                             "or the number of residues (%u) "%(len(colors),
                                                                len(residxs_as_fragments),
                                                                len(_np.hstack(residxs_as_fragments))))
    elif isinstance(colors, dict):
        assert len(colors) == len(residxs_as_fragments), (len(colors), len(residxs_as_fragments))
        col_list = [[val] * len(iseg) for val, iseg in zip(colors.values(), residxs_as_fragments)]
        col_list = _np.array([item for sublist in col_list for item in sublist])
        #col_list = _np.hstack([[val] * len(iseg) for val, iseg in zip(colors.values(), residxs_as_fragments)])
    else:
        raise Exception

    if alpha!=1:
        col_list = [_make_color_transparent(_to_rgb(col), alpha) for col in col_list]

    return col_list

def draw_arcs(fragments, iax, colors=None,
              lw=1,
              center=0, r=1, angle_offset=0, **cartify_kwargs):

    if colors is None:
        colors = ["k"]*len(fragments)
    angles = cartify_fragments(fragments,r=r, return_angles=True, **cartify_kwargs)[1]*180/_np.pi
    from mdciao.utils.lists import re_warp
    angles = re_warp(angles,[len(ifrag) for ifrag in fragments])
    from matplotlib.patches import Arc as _Arc
    for ii, iang in enumerate(angles):
        iarc = _Arc(center,
                    width=r,
                    height=r,
                    angle=angle_offset,
                    theta1=iang[-1],
                    theta2=iang[0],
                    lw=lw,
                    color=colors[ii],
                    )
        iax.add_artist(iarc)

def should_this_residue_pair_get_a_curve(
                                         fragments,
                                         mute_fragments=None,
                                         anchor_fragments=None,
                                         top=None,
                                         select_residxs=None,
                                         exclude_neighbors=0,
                                         ):
    r"""
    lambda for selecting and/or muting residue pairs

    A residue pair gets muted when at least one condition is met,
    even if another would "un-mute" it.

    Parameters
    ----------
    fragments : iterable if iterable of ints
        How residue idxs are divided into fragments.
        Any residue not contained here will
        not get a curve.
    mute_fragments : iterable of ints, default is None
        Idxs of fragments to be muted: any residue pair
        containing at least one residue in these fragments
        will be muted.
        Cannot be provided simultaneously with :obj:`anchor_fragments`
    anchor_fragments : iterable of ints, default is None
        Complementary of :obj:`mute_fragments`: any residue
        pair not containing at least one residue
        in these fragments will be muted.
        Cannot be provided simultaneously with :obj:`mute_fragments`
    select_residxs : list of idxs
        When provided, any pair *not* containing at least
        one of these residues will be muted
    exclude_neighbors : int, default is 0
        pairs with neighbors up to :obj:`exclude_neighbors`
        will be excluded. When no :obj:`top` is provided,
        the neighborhood condition is computed by
        substraction of the idxs themselves, i.e.
        [20-21] are 1-neighbors regardless of
        whether there is a bond or not connecting
        them
    top : :obj:`mdtraj.Topology`
        For implementing :obj:`exclude_neighbors` with
        topology (=bond) information, see :obj:`bonded_neighborlist_from_top`

    Returns
    -------
    lambda : pair -> boolean
        One can apply this lambda to any residue pair and it will return
        False if any of the muting conditions apply, otherwise True
    """

    is_pair_not_muted_bc_missing_in_fragment_def = lambda pair : len(_np.intersect1d(pair,_np.hstack(fragments)))==2

    # Condition residues
    if select_residxs is not None:
        is_pair_not_muted_bc_direct_selection = lambda pair: len(_np.intersect1d(pair, select_residxs)) > 0
    else:
        is_pair_not_muted_bc_direct_selection = lambda pair: True

    # Condition in anchor segment or in muted segment
    is_pair_not_muted_bc_anchor_and_mute_fragments = \
        fragment_selection_parser(fragments,
                                  hide_fragments=mute_fragments,
                                  only_show_fragments=anchor_fragments)

    # Condition not nearest neighbors
    if top is None:
        is_pair_not_muted_bc_nearest_neighbors = lambda pair: _np.abs(pair[0] - pair[1]) > exclude_neighbors
    else:
        nearest_n_neighbor_list = bonded_neighborlist_from_top(top, exclude_neighbors)
        is_pair_not_muted_bc_nearest_neighbors = lambda pair: pair[1] not in nearest_n_neighbor_list[pair[0]]

    lambda_out = lambda res_pair : is_pair_not_muted_bc_nearest_neighbors(res_pair) and \
                                   is_pair_not_muted_bc_anchor_and_mute_fragments(res_pair) and \
                                   is_pair_not_muted_bc_direct_selection(res_pair) and \
                                   is_pair_not_muted_bc_missing_in_fragment_def(res_pair)

    return lambda_out

def add_fragment_labels(fragments,
                        fragment_names,
                        iax,
                        angle_offset=0,
                        padding=[0,0,0],
                        fontsize=5,
                        center=[0,0],
                        r=1.0,
                        xy=None):
    r"""
    Add fragment labels to a flareplot

    Very similar to :obj:`add_residue_labels` but does not
    "radiate" the labels, it puts them in the angular
    center of the residues contained in each fragment

    Parameters
    ----------
    fragments : iterable if iterables of ints
        The fragment definitions
    fragment_names  :iterable of strs, len(fragments)
        The fragment names
    iax : :obj:`~matplotlib.axes.Axes`
        The axis to write on
    angle_offset : scalar, scalar default is 0
        Where the circle starts, in degrees. 0 means 3 o'clock,
        90 12 o'clock etc. It's the phi of polar coordinates
    padding : list, default is [0,0,0]
        * first integer : Put this many empty positions before the first dot
        * second integer: Put this many empty positions between fragments
        * third integer : Put this many empty positions after the last dot
    fontsize : float, default is 5
        The fontsize in which the fragment names are
        written
    center : pair for floats, default is (0,0)
        The center of the flareplot
    r : scalar, default is 1
        The radius at which the labels will be put
    xy : list or 2D np.ndarray, None
        The positions of the members of each fragment,
        has to be of len==_np.hstack(fragments).
        If None, then positions will be computed by
        :obj:`cartify_fragments` using :obj:`r`,
        :obj:`angle_offset`, and :obj:`padding`.

    Returns
    -------
    fragment_labels : list of the :obj:`~matplotlib.text.Text` objects

    """
    if xy is None:
        _xy = cartify_fragments(fragments,
                                r=r,
                                angle_offset=angle_offset,
                                padding=padding)
        _xy += center
    else:
        assert len(xy)==len(_np.hstack(fragments))
        _xy = xy
    fragment_labels = []
    residx2xyidx = _re_warp(_np.arange(len(_np.hstack(fragments))), [len(ifrag) for ifrag in fragments])
    for seg_idxs, iname in zip(residx2xyidx, fragment_names):
        xseg, yseg = _xy[seg_idxs].mean(0) - center
        rho, phi = cart2pol(xseg, yseg)
        xseg, yseg = pol2cart(r , phi) + _np.array(center)

        ha="center"
        if yseg<0:
            va="top"
        else:
            va="bottom"
        iang = phi + _np.pi / 2
        if _np.cos(iang) < 0:
            iang = iang + _np.pi

        fragment_labels.append(iax.text(xseg, yseg, iname,
                                        ha=ha,
                                        va=va,
                                        bbox={"boxstyle": "square,pad=0.0", "fc": "none", "ec": "none", "alpha": .05},
                                        # we need this transparent Fancybox to check for overlaps
                                        fontsize=fontsize,
                                        rotation_mode="anchor",
                                        rotation=_np.rad2deg(iang)))

    return fragment_labels

def add_parent_labels(kwargs_freqs2flare, flareplot_attrs, kwargs_pffp):
    r"""
    Add the outer ring of "parent" fragment names to a flareplot that's using subdomains

    Note
    ----
    This is an ad-hoc **after** ContactGroup.plot_freqs_as_flareplot
    has called freqs2flare. Currently, I don't want to change
    that freqs2flare's signature. Hence, there's three different
    dictionaries organizing the input. This can/will change
    in the future

    Parameters
    ----------
    kwargs_freqs2flare : dict
        The dictionary of kwargs computed inside
        plot_freqs_as_flareplot  by using
        _dataframe2flarekwargs and
        then sent to freqs2flare
    flareplot_attrs : dict
        The dictionary of flareplot attributes
        produced by freqs2flare.
    kwargs_pffp : dict
        The optional arguments passed
        initially to plot_freqs_as_flareplot.
        It has to contain the keys "fragments"
        and "fragment_names"
    Returns
    -------
    flareplot_attrs : dict
        The same input dictionary with
        one new key "parent_labels"
    """
    if None in [kwargs_pffp.get("fragments"), kwargs_pffp["fragment_names"]]:
        pass
    else:
        not_passed_frags = list(set(kwargs_pffp["fragment_names"]).difference(kwargs_freqs2flare["fragment_names"]))
        # some original parents weren't passed to freqs2flare. This might be
        # 1) because the were excluded completely
        # 2) they were broken down into subfragments
        if not_passed_frags:
            # Let's find out if if we're in number 2)
            # First, we need to associate actual Text-objects
            # with actually passed kwargs_freqs2flare["fragments"]
            # since even some kwargs_freqs2flare["fragments"] might've
            # been excluded by freqs2flare because of sparse-type arguments
            plotted_fragTexts_by_input_idx = _defdict(list)
            for ff, fn in enumerate(kwargs_freqs2flare["fragment_names"]):
                for tt, txt in enumerate(flareplot_attrs["fragment_labels"]):
                    if txt.get_text() == _replace4latex(fn):  # The text object has been latexified already
                        plotted_fragTexts_by_input_idx[ff].append(tt)

            # Assert uniqueness
            assert all([len(val) == 1 for val in plotted_fragTexts_by_input_idx.values()])
            plotted_fragTexts_by_input_idx = dict(plotted_fragTexts_by_input_idx)
            plotted_fragTexts_by_input_idx = {key: val[0] for key, val in plotted_fragTexts_by_input_idx.items()}
            parent_labels, parent_xy = [], []

            #Store the axis before it gets removed
            iax = flareplot_attrs["fragment_labels"][0].axes
            _, kid_by_parents = _find_parent_list(kwargs_freqs2flare["fragments"], kwargs_pffp["fragments"])
            for parent_idx, kids in kid_by_parents.items():
                present_kids_by_TextObject_idx = [plotted_fragTexts_by_input_idx[kk] for kk in kids if kk in plotted_fragTexts_by_input_idx.keys()]
                if present_kids_by_TextObject_idx:
                    present_TextObjects = [flareplot_attrs["fragment_labels"][idx] for idx in present_kids_by_TextObject_idx]
                    present_weights =     [len(kwargs_freqs2flare["fragments"][kk]) for kk in kids if kk in plotted_fragTexts_by_input_idx.keys()]
                    assert len(present_weights)==len(present_TextObjects)
                    parent_xy.append(_np.average([tt.get_position() for tt in present_TextObjects],
                                                 axis=0, weights=present_weights))
                    parent_labels.append(kwargs_pffp["fragment_names"][parent_idx])
                    if len(present_TextObjects)==1 and present_TextObjects[0].get_text()==_replace4latex(parent_labels[-1]):
                        present_TextObjects[0].remove()

            if parent_labels:
                parent_labels = add_fragment_labels([[None]]*len(parent_labels),
                                                    [_replace4latex(str(ifrag)) for ifrag in
                                                     parent_labels],
                                                    iax,
                                                    fontsize=flareplot_attrs["fragment_labels"][0].get_fontsize() * 2,
                                                    r=flareplot_attrs["r"]+flareplot_attrs["dots"][0].radius*4,
                                                    xy=_np.array(parent_xy))
                flareplot_attrs["parent_labels"] = parent_labels

    return flareplot_attrs

def change_axlims_and_resize_Texts(iax, new_r, center=[0, 0]):
    r"""
    Change the x- and y-lims of a square plot and scale the Text objects to the new limits

    In matplotlib, fontsizes of Text-objects are constant in pts.
    Changing the axis limits doesn't affect the text size, but it
    affects their position. This means that on "zooming out", the
    text objects get closer to one another, potentially causing
    overlaps. This method scales the fontsize down by the amount
    of lost distance between text-object, effectively making them
    keep their size in axis data units

    Parameters
    ----------
    iax
    new_r
    center

    Returns
    -------

    """
    old_d_x = _np.abs(_np.diff(iax.get_xlim()))
    old_d_y = _np.abs(_np.diff(iax.get_ylim()))
    assert iax.get_aspect()==1
    assert old_d_x==old_d_y
    iax.set_xlim([center[0] - new_r, center[0] + new_r])
    iax.set_ylim([center[1] - new_r, center[1] + new_r])
    new_d = _np.abs(_np.diff(iax.get_xlim()))
    [lab.set_fontsize(lab.get_fontsize() * (old_d_x / new_d)) for lab in iax.texts]

def cart2pol(x, y):
    rho = _np.sqrt(x**2 + y**2)
    phi = _np.arctan2(y, x)
    return(rho, phi)

def add_residue_labels(iax,
                       xy_labels,
                       res_idxs,
                       fontsize,
                       center=None,
                       top=None,
                       shortenAAs=True,
                       aa_offset=0,
                       colors=None,
                       highlight_residxs=None,
                       replacement_labels=None,
                       **text_kwargs
                       ):
    r"""
    Add residue names to a flareplot

    Parameters
    ----------
    iax : :obj:`~matplotlib.axes.Axes`
    res_idxs : np.ndarray
        The residue indices to use as labels
    xy_labels : 2D-np.ndarray
        x-y position of the labels
    center: pair of floats, default is None
        When provided, this is the center
        of the circle that the :obj:`xy_labels`
        are assumed to lie on. The polar rho-angle
        will be computed and passed as :obj:`rotation`
        argument to the :obj:`matplotlib.text` call,
        making residue labels "radiate"
        out of the :obj:`center`.
    fontsize : int
    top : :obj:`mtdtraj.Topology`, default is None
        Use the idxs in :obj:`res_idxs` on this
        topology to generate better labels
    shortenAAs: bool, default is True
        Use the short-code for the AAS (GLU30->E30)
    aa_offset : int, default is 0
        Add this number to the resSeq value
    colors : list of len(res_idxs), default is None,
        Individual residue colors, default is black
    replacement_labels : dict, default is None
        Input individual replacements for residue labels here,
        keyed with residue idxs
        Typical cases could be a mutated residue that you want
        to show as R38A instead of just A38, or use
        e.g. GPCR or CGN consensus labels.
    highlight_residxs : iterable of ints, default is None
        In case you don't want to construct a whole
        color list for :obj:`colors`, you can simply
        input a subset of :obj:`res_idxs` here and
        they will be shown in red.


    Returns
    -------
    labels : list of :obj:`matplotlib.text.Text` objects

    """
    assert len(res_idxs) == len(xy_labels)

    if colors is None:
        colors = ['k']*len(res_idxs)
    else:
        assert len(colors)==len(res_idxs)

    if replacement_labels is None:
        replacement_labels = {}

    labels = []
    for ii, (res_idx, ixy) in enumerate(zip(res_idxs, xy_labels)):
        ilabel = res_idx
        txtclr = colors[ii]

        if top is not None:
            if not shortenAAs:
                ilabel = top.residue(res_idx)
            else:
                idxs = top.residue(res_idx).resSeq + aa_offset
                ilabel = ("%s%u" % (top.residue(res_idx).code, idxs)).replace("None", top.residue(res_idx).name)
            if highlight_residxs is not None and res_idx in highlight_residxs:
                txtclr = "red"

        try:
            ilabel = replacement_labels[res_idx]
        except KeyError:
            pass # the residue does not have a replacement

        rotation = 0
        ha="left"
        if center is not None:
            rho, phi = cart2pol(ixy[0] - center[0], ixy[1] - center[1])
            if _np.cos(phi) < 0:
                phi = phi + _np.pi
                ha = "right"
            rotation = phi
        itxt = iax.text(ixy[0], ixy[1], '%s' % ilabel,
                        color=txtclr,
                        va="center",
                        ha=ha,
                        rotation=_np.rad2deg(rotation),
                        rotation_mode="anchor",
                        fontsize=fontsize,
                        zorder=20,
                        bbox={"boxstyle":"square,pad=0.0","fc":"none", "ec":"none", "alpha": .05},  # we need this transparent Fancybox to check for overlaps
                        **text_kwargs)
        labels.append(itxt)

    return labels


_SS2vmdcol = {'H':"purple", "E":"yellow","C":"cyan", "NA":"gray"}

def value2position_map(unique_idxs):
    r"""
    Return an dictionary v2p so that v2p[idx]=pos, where pos is the position of idx in unique_idxs

    Parameters
    ----------
    unique_idxs : iterable of ints
        No index can be repeated

    Returns
    -------
    value2pos : dict


    """
    values, positions = _np.unique(unique_idxs, return_index=True)
    assert len(values)==len(positions)
    value2pos = {key:val for key, val in zip(values, positions)}
    return value2pos

def overlappers(text_objects_1, text_objects_2):
    r"""
    Return a list of objects whose window_extents overlap in any way with any other object

    The self-overlap is ignored

    Parameters
    ----------
    text_objects_1 : list
        List of :obj:`matplotlib.text.Text` objects

    Returns
    -------
    overlappers : list
        The :obj:`matplotlib.text.Text` objects


    """

    boxes_1 = [t.get_window_extent(renderer=t.axes.figure.canvas.get_renderer()) for t in text_objects_1]
    # We could be computing the same boxes two times but makes for
    # better readability and more consistent return values
    boxes_2 = [t.get_window_extent(renderer=t.axes.figure.canvas.get_renderer()) for t in text_objects_2]
    return [t2 for t2, b2 in zip(text_objects_2,boxes_2) if any([(t1 is not t2 and b1.overlaps(b2)) for t1, b1 in zip(text_objects_1,boxes_1)])]

def overlappers_one_to_one(objects_1, objects_2,break_loop=True):
    r"""
    Check for overlaps of the i-th element of each list (no cross-pairs)


    Parameters
    ----------
    objects_1 : list
        List of objects with get_window_extent method

    objects_2 : list
        List of objects with get_window_extent method

    break_loop : bool, default is False
        Break the loop at the first overlap

    Returns
    -------
    overlappers : list
        List with the idxs of the pairs that overlap
        If break_loop is True, will contain only
        the first entry


    """

    assert len(objects_1)==len(objects_2)
    res = []
    for ii, (obj1,obj2) in enumerate(zip(objects_1,objects_2)):
        b1 = obj1.get_window_extent(renderer=obj1.axes.figure.canvas.get_renderer())
        b2 = obj2.get_window_extent(renderer=obj2.axes.figure.canvas.get_renderer())
        if b1.overlaps(b2):
            res.append(ii)
            if break_loop:
                break
    return res



def un_overlap_via_fontsize(text_objects, fac=.95, maxiter=50):
    r"""
    Iteratively (up to n_max) reduce the fontsize by :obj:`fac`
    until the text objects do not overlap anymore

    Parameters
    ----------
    text_objects : list
        List of :obj:`matplotlib.text.Text` objects
    fac : float, default is .95
        fontsize_ii_+1 = fontsize_ii * fac
    maxiter: int, default is 50
        Stop after these many iterations

    Returns
    -------

    """
    counter = 0
    while len(text_objects)>0 and any_overlap_via_FancyBoxPach(text_objects, text_objects) and counter < maxiter:
        [t1.set_size(t1.get_size() * fac) for t1 in text_objects]
        counter += 1

#TODO RENAME
def _parse_residue_and_fragments(res_idxs_pairs, sparse_residues=False,
                                 sparse_fragments=False, fragments=None, top=None,
                                 anchor_fragments=None, mute_fragments=None):
    r""" Decide what residues the user wants to get a dot.

    Typically, the output of this function will get parsed to :obj:`circle_plot_residues`

    Note
    ----
    I've outsourced this decision-making here because it's not clear to me yet
    what the best interplay in sparse_residues vs fragment could be

    Parameters
    ----------
    res_idxs_pairs : np.ndarray
        (N,2)-array with N residue pairs
    sparse_residues : boolean or iterable of ints, default is False
        * If False, return an array (0,np.max(res_idxs_pairs)+1)
        * If True,  return only np.unique(res_idxs_pairs)
        * If iterable of ints, override everything and return these indices
    sparse_fragments : bool, default is False
        If True, return all residues of the fragments where
        at least one residue is present in res_idxs_pairs
        An error will be raised if sparse_fragment is True
        and sparse_residues isn't False
    fragments : list, default is None
        List of integer lists representing how the residues
        are grouped together in fragments. Their union
        must always be a superset of whatever set resulted
        of the chosen :obj:`sparse_residues` option. If sparse_residues is False,
        residues present here will be added to the output even
        if they don't contain any relevant residue
    top : :obj:`~mdtraj.Topology`, default is None
        Only used if :obj:`fragments` is None
        and :obj:`sparse_residues` is False. Then, if
        top is None, dots will be assigned only up
        to max(res_idxs), and if top is passed, up to
         :obj:`top.n_residues`
    anchor_fragments : list, default is None
        In case this was passed as not None initially
        a re-indexing needs to happen here
    mute_fragments : list, default is None
        In case this was passed as not None initially
        a re-indexing needs to happen here

    Returns
    -------
    residues_as_fragments, anchor_fragments, mute_fragments
    """
    res_idxs = _np.unique(res_idxs_pairs)
    if sparse_fragments:
        assert sparse_residues is False
    if fragments is None:
        if isinstance(sparse_residues, bool):
            if sparse_residues:
                residues_as_fragments=[res_idxs]
            else:
                if top is None:
                    residues_as_fragments=[_np.arange(res_idxs[-1] + 1)]
                else:
                    residues_as_fragments=[_np.arange(top.n_residues)]
        else:
            residues_as_fragments=[sparse_residues]
    else:
        # Checks
        _no_intersect(fragments, word="fragments")
        if not set(res_idxs).issubset(_np.hstack(fragments)) and isinstance(sparse_residues, bool):
            raise ValueError("The input fragments do not contain all residues residx_array, " \
            "their set difference is %s" % sorted(set(res_idxs).difference(_np.hstack(fragments))))

        if isinstance(sparse_residues, bool):
            if sparse_residues:
                residues_as_fragments = [_np.intersect1d(ifrag,res_idxs) for ifrag in fragments]
            else:
                if sparse_fragments:
                    residues_as_fragments = [ifrag for ifrag in fragments if len(_np.intersect1d(ifrag,res_idxs))>0]
                else:
                    residues_as_fragments = fragments
        else:
            residues_as_fragments = [_np.intersect1d(ifrag, sparse_residues) for ifrag in fragments]

        residues_as_fragments = [ifrag for ifrag in residues_as_fragments if len(ifrag) > 0]

    re_indexed = []
    #TODO a lot of re-indexing or selections elswhere using some variation of "intersect_with"
    #try to refactor in one general method or many methods grouped together
    for am_frags in [anchor_fragments, mute_fragments]:
        if am_frags is not None:
            assert fragments is not None
            intersect_with = _np.hstack([fragments[ii] for ii in am_frags])
            am_frags = [ii for ii, ifrag in enumerate(residues_as_fragments) if len(_np.intersect1d(ifrag,intersect_with))>0]
            if len(am_frags)==0:
                am_frags = None
        else:
            pass
        re_indexed.append(am_frags)
    anchor_fragments, mute_fragments = re_indexed

    return residues_as_fragments, anchor_fragments, mute_fragments


def fontsize_get(iax):
    r"""
    Scan text elements and return a dictionary with fontsizes
    Parameters
    ----------
    iax : :obj:`~matplotlib.axes.Axes`

    Returns
    -------
    sizes : dict
        A dictionary with two lists,
        keyed with "n_polygons" and "other"
        * "n_polygons" contains a list in ascending
        order with the fontsizes of text elements
        of which there are as many as CirclePolygons
        This fact is a strong hint that those text
        elements correspond 1st to the residue-label texts
        and 2nd to the residue-SS-label texts
        * "other" contains all other fontsizes in
        :obj:`iax`, most likely belonging to the
        fragment-label texts
    """
    n_polygons = len([obj for obj in iax.findobj(_CP)])
    sizes = {"n_polygons": [],
             "other": []}

    c = _Counter([_np.round(txt.get_fontsize(), 2) for txt in iax.texts])

    for key, val in c.items():
        sizes[{True: "n_polygons",
               False: "other"}[val == n_polygons]].append(key)

    return {key: sorted(val) for key, val in sizes.items()}


def fontsize_apply(axA, axB):
    r"""
    Map the fontsizes of axA to the text-objects of axB

    Internally, :obj:`fontsize_get` is used
    Parameters
    ----------
    axA : :obj:`~matplotlib.axes.Axes`
    axB : :obj:`~matplotlib.axes.Axes`

    Returns
    -------

    """
    sizes1 = fontsize_get(axA)
    sizes2 = fontsize_get(axB)

    for key, val in sizes2.items():
        assert len(sizes1[key]) == len(val)

    fs2type = {}
    for key, val in sizes2.items():
        for ii, vv in enumerate(val):
            fs2type[vv] = sizes1[key][ii]

    for txt in axB.texts:
        fs = _np.round(txt.get_fontsize(), 2)
        txt.set_fontsize(fs2type[fs])

def add_aura(xy, aura, iax, r=1, fragment_lenghts=None, width=.10, subtract_baseline=True,
             colors=None,
             lines=False):
    r"""
    Create "auras", i.e. outer rings of the flareplot representing any scalar array

    Parameters
    ----------
    xy : np.ndarray of shape (N,2)
    aura : iterable of positive scalars
        The value to be plotted, e.g.
        conservation degree, RMSF,
        SASA, whatever.
        The indices are indices of
        :obj:`xy`, s.t. the have to
        be in the interval [0,len(xy)]
        If :obj:`xy` has been "sparsed"
        somehow, :obj:`aura` needs also
        to be sparsed
    r : float, default is 1
        Radius of the aura's baseline
    iax : :obj:`~matplotlib.axes.Axes`
        The axes containing the flareplot
    width : .10
        The width of the aura, in units of :obj:`r`
        The maximum aura value will always appear
        at r+r*width
    fragment_lenghts : iterable of ints, default is None
        How the :obj:`xy` is divided into fragments.
        if None, defaults to [len(xy)]
    subtract_baseline : bool, default is True
        Subtract the baseline of the :obj:`aura`.
        This highlights differences between
        high-values and low values,
        but makes the minimum values appear as zero.
        If False, then all values of the aura get compressed
        to the range [r,r+r*width], which can
        lead to poor resolution. For example, if
        you are plotting conservation degree in percent and
        all values are between 85 and 100%, not subtracting
        the baseline will make it hard to distinguish anything,
        all values will appear as "high".
    colors : anything
    lines : bool, default is False
        Use lines instead of polygons to represent the aura

    Returns
    -------
    artists : list
        Contains the  artist objects, either
        :obj:`~matplotlib.patches.Polygon` or
        :obj:`~matplotlib.collections.LineCollection`
    rmax : float
        The new maximum radius, i.e. (1+width)*r
    """

    if fragment_lenghts is None:
        fragment_lenghts = [len(xy)]
    assert len(aura) == len(xy) == sum(fragment_lenghts)
    aura = _np.array(aura,dtype=float)
    if any([a<0 for a in aura]):
        raise NotImplementedError("Negative aura-values not implemented yet")

    aura /= aura.max()
    if subtract_baseline:
        aura -= aura.min()
        aura /= aura.max()
    assert  aura.max()==1

    aura *= r * width

    col_list = col_list_from_input_and_fragments(colors, [_np.arange(l) for l in fragment_lenghts],alpha=.80)

    artists = []
    aura_iter = iter(aura)
    for jj, ixy in enumerate(_re_warp(xy, fragment_lenghts)):
        base, top, segments = [], [], []
        for ii, (x, y) in enumerate(ixy):
            theta = _np.arctan2(y, x)
            iaura = next(aura_iter)
            #theta_deg = theta / _np.pi * 180
            # print(theta)
            # plt.text(x,y,"%s %u"%(ii,theta_deg),fontsize=20)
            start = [_np.cos(theta) * r, _np.sin(theta) * r]
            end = [_np.cos(theta) * (r + iaura), _np.sin(theta) * (r + iaura)]
            base.append(start)
            top.append(end)
            segments.append([start, end])
        base = _np.array(base)
        top = _np.array(top)
        poly = _np.vstack((base, top[::-1]))
        # iax.scatter(base[:,0],base[:,1],color="r",zorder=100)
        # iax.scatter(top[:,0],top[:,1],color="g",zorder=100)
        if lines:
            artists.append(_LCol(segments, color=col_list[jj]))
        else:
            artists.append(_PG(poly, alpha=.80, color=col_list[jj],lw=0))

        iax.add_artist(artists[-1])
    return artists, r+aura.max()

def coarse_grain_freqs_by_frag(freqs, res_idxs_pairs, fragments,
                               check_if_subset=True):
    r"""
    Coarse-grain per-residue frequencies into a per-fragment contact-matrix

    The contact matrix gets symmetrized regardless of the input residue indices

    Parameters
    ----------
    freqs : iterable of floats
        The frequencies
    res_idxs_pairs : iterable of pairs
        The pairs
    fragments: iterable of iterables
        The fragments. Each individual
        residue of :obj:`res_index_pairs` can
        appear at most in one (and only one) fragment
        TODO this might not be true anymore,
        since self_interface can be True
    check_if_subset : bool, default is True
        Check whether all idxs in
        :obj:`res_idxs_pairs` belong
        to some fragment of :obj:`fragments`
        and raise an error if they dont.
        Bypassing this check  may lead
        to the residues missing from
        :obj:`fragments` not contributing
        to the freqs of the coarse-grained matrix

    Returns
    -------
    mat : np.ndarray
        Square, symmetric matrix of shape (len(fragments),len(fragments))
        TODO think about returning NaNs instead of zeros
        for the elements not present in :obj:`res_idxs_pairs`
    """

    parents, children = assign_fragments(_np.unique(res_idxs_pairs), fragments, raise_on_missing=check_if_subset)

    res_max = _np.max([_np.hstack(fragments).max(), _np.unique(res_idxs_pairs).max()])
    idx2frag = _np.zeros(res_max + 1)
    idx2frag[:] = _np.nan
    idx2frag[children] = parents
    frag_pairs = idx2frag[_np.array(res_idxs_pairs)].astype(int)
    mat_CG = _np.zeros((len(fragments), len(fragments)))
    for rp, fp, freq in zip(res_idxs_pairs, frag_pairs, freqs):
        if len(_np.intersect1d(rp, children)) == 2:
            mat_CG[fp[0], fp[1]] += freq
            mat_CG[fp[1], fp[0]] += freq

    mat_CG[_np.arange(len(fragments)), _np.arange(len(fragments))] /= 2 # the self-contacts need to be div 2

    return mat_CG

def sparsify_sym_matrix(mat, eps=1e-2):
    r"""
    Return a sparse matrix

    Checks of > eps are done element-wise,
    for entire row/column sum checks,
    see :obj: sparsify_sym_matrix_by_row_sum

    Parameters
    ----------
    mat : 2D np.ndarray of shape(N,N)
        The symmetric matrix
    eps : float, default is 1e-2
        Elements of :obj:`mat` are
        considered non-zero if > eps

    Returns
    -------
    mat, non_zeros
    """
    assert (mat==mat.T).all() #checks for squareness and symmetry
    non_zeros = _np.flatnonzero(_np.any(mat > eps, axis=1))
    mat = mat[non_zeros, :][:, non_zeros]
    return mat, non_zeros

def sparsify_sym_matrix_by_row_sum(mat, eps=1e-2):
    r"""
    Return a sparse matrix

    Parameters
    ----------
    mat : 2D np.ndarray of shape(N,N)
        The symmetric matrix
    eps : float, default is 1e-2
        Rows/Columns Elements of :obj:`mat` are
        considered non-zero row.sum() > eps

    Returns
    -------
    mat, non_zeros
    """
    assert (mat==mat.T).all() #checks for squareness and symmetry
    non_zeros = _np.flatnonzero(mat.sum(1)>eps)
    zeros = _np.flatnonzero(mat.sum(1)<=eps)
    discarded = mat.sum(1)[zeros].sum()
    mat = mat[non_zeros, :][:, non_zeros]
    return mat, non_zeros, discarded


r"""
Functions for manipulating strings and dictionaries, also a bit of IO.


.. currentmodule:: mdciao.utils.str_and_dict


Functions
=========

.. autosummary::
   :toctree: generated/

"""
from glob import glob as _glob
import numpy as _np
import mdtraj as _md
from .lists import re_warp, contiguous_ranges as _cranges
from fnmatch import fnmatch as _fnmatch
from pandas import read_excel as _read_excel
from os import path as _path, listdir as _ls
import re as _re
from collections import defaultdict as _defdict
from natsort import natsorted as _natsorted

tunit2tunit = {"ps":  {"ps": 1, "ns": 1e-3, "mus": 1e-6, "ms":1e-9},
                "ns":  {"ps": 1e3, "ns": 1,    "mus": 1e-3, "ms":1e-6},
                "mus": {"ps": 1e6, "ns": 1e3,  "mus": 1,    "ms":1e-3},
                "ms":  {"ps": 1e9, "ns": 1e6,  "mus": 1e3,  "ms":1},
               }

def get_sorted_trajectories(trajectories):
    r"""
    Common parser for something that can be interpreted as a trajectory

    Parameters
    ----------
    trajectories: can be one of these things:
        - pattern, e.g. "*.ext"
        - one string containing a filename
        - list of filenames
        - one :obj:`mdtraj.Trajectory` object
        - list of :obj:`mdtraj.Trajectory` objects

    Returns
    -------
        - for an input pattern, sorted trajectory filenames that match that pattern
        - for filename, one list containing that filename
        - for a list of filenames, a sorted list of filenames
        - for one :obj:`mdtraj.Trajectory` object, a list containing that object
        - list of :obj:`mdtraj.Trajectory` objects (i.e. does nothing)


    """
    if isinstance(trajectories,str):
        _trajectories = _glob(trajectories)
        if len(_trajectories)==0:
            raise FileNotFoundError("Couldn't find (or pattern-match) anything to '%s'.\n"
                                    "ls $CWD[%s]:\n%s:"%(trajectories,
                                                         _path.abspath(_path.curdir),
                                                         "\n".join(_ls(_path.curdir))))
        else:
            trajectories=_trajectories

    if isinstance(trajectories[0],str):
        xtcs = _natsorted(trajectories)
    elif isinstance(trajectories, _md.Trajectory):
        xtcs = [trajectories]
    else:
        assert all([isinstance(itraj, _md.Trajectory) for itraj in trajectories])
        xtcs = trajectories

    return xtcs

def inform_about_trajectories(trajectories, only_show_first_and_last=False):
    r"""
    Return a string that informs about the trajectories

    Parameters
    ----------
    trajectories: list of strings or :obj:`mdtraj.Trajectory` objects


    Returns
    -------
    listed_str : a string with the trajectory names separated by newlines

    """
    ntraj = len(trajectories)
    assert isinstance(trajectories, list), "input has to be a list"
    if isinstance(only_show_first_and_last,int) and only_show_first_and_last*2<ntraj:
        n = only_show_first_and_last
        _trajectories =trajectories[:n]+["...[long list: omitted %u items]..."%(ntraj-2*n)]+trajectories[-n:]

    else:
        _trajectories = trajectories
    return "\n".join([str(itraj) for itraj in _trajectories])

def replace_w_dict(input_str, exp_rep_dict):
    r"""
    Sequentially perform string replacements on a string using a dictionary

    Parameters
    ----------
    input_str: str
    exp_rep_dict: dictionary
        keys are expressions that will be replaced with values, i.e.
        key = key.replace(key1, val1) for key1, val1 etc

    Returns
    -------
    key

    """
    for pat, exp in exp_rep_dict.items():
        input_str = input_str.replace(pat, exp)
    return input_str

def delete_exp_in_keys(idict, exp, sep="-"):
    r"""
    Assuming the keys in the dictionary are formed by two segments
    joined by a separator, e.g. "GLU30-ARG40", deletes the segment
    containing the input expression, :obj:`exp`

    Will fail if not all keys have the expression to be deleted

    Parameters
    ----------
    idict: dictionary
    exp: str
    sep: str, default is "-",

    Returns
    -------
    dict:
        dictionary with the same values but the keys lack the
        segment containing :obj:`exp`

    dhk : list
        List with the deleted half-keys
    """

    out_dict = {}
    deleted_half_keys=[]
    for names, val in idict.items():
        new_name, dhk = delete_pattern_in_ctc_label(exp,names,sep)
        deleted_half_keys.extend(dhk)
        out_dict[new_name]=val
    return out_dict,deleted_half_keys

def delete_pattern_in_ctc_label(pattern, label, sep):
    new_name = [name for name in splitlabel(label, sep) if pattern not in name]
    deleted_half_keys = [name for name in splitlabel(label,sep) if pattern in name]
    assert len(new_name) == 1, (new_name, pattern)
    return new_name[0], deleted_half_keys

# Order key alphabetically using the separator_key
def order_key(key, sep):
    split_key = splitlabel(key,sep)
    return sep.join([split_key[ii] for ii in _np.argsort(split_key)])

def unify_freq_dicts(freqs,
                     exclude=None,
                     key_separator="-",
                     replacement_dict=None,
                     defrag=None,
                     per_residue=False,
                     is_freq=True,
                     val_missing=0,
                     verbose=True
                     ):
    r"""
    Provided with a dictionary of dictionaries, returns an equivalent,
    key-unified dictionary where all sub-dictionaries share their keys,
    putting zeroes where keys where absent originally.

    Use :obj:`key_separator` for "GLU30-LY40" == "LYS40-GLU30" to be True

    Parameters
    ----------
    freqs:  dictionary of dictionaries, e.g.:
        {A:{key1:valA1, key2:valA2, key3:valA3},
         B:{            key2:valB2, key3:valB3}}

    key_separator: str, default is "-"
        Specify how residues are separated in the contact
        label, eg. "GLU30-LYS40".
        With this knowledge, the method can split the label
        before comparison so that "GLU30-LYS40" is considered
        equal to "LYS40-GLU30". Use "", "none" or None to differentiate.
        It will also be passed to :obj:`defrag_key` in case
        :obj:`defrag` is not None.

    exclude: list, default is None
         keys containing these strings will be excluded.
         NOTE: This is not implemented yet, will raise an error

    replacement_dict: dict, default is {}
        all keys/strings will be subjected to replacements following this
        dictionary, st. "GLH30" is "GLU30" if replacement_dict is {"GLH":"GLU"}
        This way mutations and or indexing can be accounted for in different setups
    defrag : char, default is None
        If a char is given, "@", anything after that character in the labels
        will be consider fragment information and ignored. This is only recommended
        for advanced users, usually the fragment information helps keep track
        of residue names in complex topologies:
            R201@frag1 and R201@frag3 will both be "R201"
    per_residue : bool, default is False
        Aggregate interactions to their residues
    is_freq : bool, default is True
        If the dictionaries actually
        contain frequencies or not.
        If not, some checks are omitted
    val_missing : anything, default is 0
        What value to assign to the
        missing keys (TODO check the name of this in pandas)
    verbose : bool, default is True
        Be verbose


    Returns
    -------
    unified_dict: dictionary
        A dictionary  of dictionaries sharing keys:
       {A:{key1:valA1, key2:valA2, key3:valA3},
        B:{key1:0,     key2:valB2, key3:valB3}}
    """

    # Create a copy
    freqs_work = {key : {key2: val for key2, val in idict.items()} for key, idict in freqs.items()}

    # Enforce replacement dictionaries
    if replacement_dict is not None:
        freqs_work = {key: {replace_w_dict(key2, replacement_dict): val2 for key2, val2 in val.items()} for key, val in
                      freqs_work.items()}

    #Re - ordered keys if needed
    if str(key_separator).lower() != "none" and len(key_separator) > 0:
        freqs_work = {key : {order_key(key2, key_separator): val2 for key2, val2 in idict.items()} for key, idict in freqs_work.items()}

    if defrag is not None:
        freqs_work = {key:{defrag_key(key2, defrag):val2 for key2, val2 in val.items()} for key, val in freqs_work.items()}

    if per_residue:
        freqs_work = {key:sum_dict_per_residue(val, key_separator) for key, val in freqs_work.items()}

    # Perform the difference operations
    not_shared = []
    shared = []
    for idict1 in freqs_work.values():
        for idict2 in freqs_work.values():
            if not idict1 is idict2:
                not_shared += list(set(idict1.keys()).difference(idict2.keys()))
                shared += list(set(idict1.keys()).intersection(idict2.keys()))

    shared = list(_np.unique(shared))
    not_shared = list(_np.unique(not_shared))
    all_keys = shared + not_shared
    # Prune keys we're not interested in
    excluded = []
    if exclude is not None:
        raise NotImplementedError("This feature not yet implemented")
        """
        assert isinstance(exclude,list)
        print("Excluding")
        for ikey, ifreq in freqs_work.items():
            # IDK I had this condition here, i think it is more intuitive if
            # they are removed regardless if shared or not
            #for key in shared:
            for key in list(ifreq.keys()):
                for pat in exclude:
                    if pat in key:
                        ifreq.pop(key)
                        print("%s from %s" % (key, ikey))
                        print(ifreq.keys())
                        excluded.append(key)
                        #all_keys = [ak for ak in all_keys if ak != key]
        """

    # Set the non shared keys to zero
    for ikey, ifreq in freqs_work.items():
        for key in all_keys:
            if key not in ifreq.keys():
                ifreq[key] = val_missing

    if len(not_shared)>0 and is_freq and verbose:
        print("These interactions are not shared:\n%s" % (', '.join(not_shared)))
        print("Their cumulative ctc freq is %3.2f. " % _np.sum(
            [[ifreq[key] for ifreq in freqs_work.values()] for key in not_shared]))

    return freqs_work

def average_freq_dict(freqs,
                      weights=None,
                      **unify_freq_dicts_kwargs
                     ):
    r"""
    Average frequencies (or anything) over dictionaries.

    Typically, the input :obj:`freqs` are keyed first by system,
    then by contact label, e.g. {"T300":{"GDP-R201":1.0},
                                "T320":{"GDP-R201":.25},
                                "MUT":{"GDP-L201":25}}

    The input data need not be unified, the method calls
    :obj:`unify_freq_dicts` internally. In the example above
    you have to call it with the arg replacement_dict={"L201:R201"}
    so tha it can understand that mutation when unifying


    Parameters
    ----------
    freqs : dict of dicts
        The dictionaries containing frequence dictionaries,

    weights : dict, default is None
        relative weights of each dictionary

    unify_freq_dicts_kwargs

    Returns
    -------
    averaged_dict : dict
        an averaged dictionary keyed only with the


    """
    freqs_work = unify_freq_dicts(freqs,**unify_freq_dicts_kwargs)

    sys_keys = list(freqs_work.keys())
    frq_keys = list(freqs_work[sys_keys[0]].keys())
    averaged_dict = {}
    if weights is None:
        weights = {key:1 for key in sys_keys}
    for fk in frq_keys:
        averaged_dict[fk] = _np.average([freqs_work[isys][fk] for isys in sys_keys],
                                        weights=[weights[isys] for isys in sys_keys])

    return averaged_dict

def sum_dict_per_residue(idict, sep):
    r"""Return a "per-residue" sum of values from a "per-residue-pair" keyed dictionary

    Note:
    There is a closely related method in :obj:`mdciao.contacts.ContactGroup`
    that allows to query the freqs from the object already aggregated
    by residue. This is for when the object is either not accessible, e.g.
    because the freqs were loaded from a file

    Parameters
    ----------
    idict : dict
        Keyed with contact labels like "res1@frag1-res2@3.50" etc
    sep : char
        Character that separates fragments in the label

    Returns
    -------
    aggr : dict
        keyed with "res1@frag1" etc

    """
    out_dict = _defdict(list)
    for key, freq in idict.items():
        key1, key2 = splitlabel(key, sep) #This will fail if sep is not in key or sep does not separate in two
        out_dict[key1].append(freq)
        out_dict[key2].append(freq)
    return {key:_np.sum(val) for key, val in out_dict.items()}

def sort_dict_by_asc_values(idict, reverse=False):
    r""" Sort a dictionary by values

    Parameters
    ----------
    idict : dict
        Input dictionary
    reverse : bool, default is False
        Reverse the sorting order,
        i.e. sort by ascending order
        of values

    Returns
    -------
    odict : dict
        Indict sorted with its keys
         sorted by its values
    """
    return {key: val for key, val in sorted(idict.items(), key=lambda item: item[1], reverse=reverse)}

def lexsort_ctc_labels(ctc_labels, reverse=False, columns=[0,1], sep="-") -> tuple:
    r"""
    Sort contact-labels in ascending order of resSeq using both columns

    Wraps around :obj:`numpy.lexsort` with some string handling

    It will also work with contact-labels consisting of only one residue,
    e.g. in the cases where the "anchor" has been deleted or the frequencies
    have been aggregated to per-residue frequencies

    >>> labels = ["ALA30@3.50-GLU50",
    >>>           "HIS28-GLU50",
    >>>           "ALA30-GLU20"]
    >>> sorted_labels, order = mdciao.utils.str_and_dict.lexsort_ctc_labels(labels)
    >>> sorted_ctc_labels
    >>> ['HIS28-GLU50',
    >>>  'ALA30-GLU20',
    >>>  'ALA30@3.50-GLU50']

    Parameters
    ----------
    ctc_labels : list of np.ndarray
        Strings describing the contact
        residues. It can contain also
        fragment information, which
        will be ignored when sorting
        but returned in :obj:`sorted_ctc_labels`
    reverse : bool, default is False
        If True, sort in descending
        order, instead of ascending
    columns : list
        The order of the columns,
        e.g. [0,1] means sort first
        by first column (idx 0),
        then by second column (idx 1).
    sep : char, default is "-"
        The character to use
        when separating the
        contact label into both residues

    Returns
    -------
    order : 1D np.ndarray
        The indices of :obj:`ctc_labels` that
        sort it into :obj:`sorted_ctc_labels`
    sorted_ctc_labels : list
        The sorted contact labels
    """
    resSeqs = _np.vstack(
        [[intblocks_in_str(pp)[0] for pp in splitlabel(lab,sep)] for lab in ctc_labels])

    if resSeqs.shape[1]==1:
        order = _np.argsort(resSeqs.squeeze())
    else:
        order = _np.lexsort([resSeqs[:, columns[1]], resSeqs[:, columns[0]]])
    if reverse:
        order=order[::-1]
    sorted_ctc_labels = [ctc_labels[ii] for ii in order]
    return sorted_ctc_labels, order

def freq_file2dict(ifile, defrag=None):
    r"""
    Read a file containing the frequencies ("freq") and labels ("label")
    of pre-computed contacts

    Parameters
    ----------
    ifile : str
        Path to file, can be a .xlsx, .dat, .txt

    defrag : str, default is None
        If passed a string, e.g "@", the fragment information
        of the contact label will be deleted upon reading,
        so that R131@frag1 becomes R131. This is done
        by calling :obj:`defrag_key` internally

    Returns
    -------
    dict : keyed by labels and valued with frequencies, e.g .{"0-1":.3, "0-2":.1}

    """
    ext = _path.splitext(ifile)[-1]
    if ext.lower() == ".xlsx":
        df = _read_excel(ifile, engine="openpyxl")
        if "freq" in df.keys() and "label" in df.keys():
            res = {key: val for key, val in zip(df["label"].values, df["freq"].values)}
        else:
            row_lab, col_lab = _np.argwhere(df.values == "label").squeeze()
            row_freq, col_freq = _np.argwhere(df.values == "freq").squeeze()
            assert row_lab == row_freq,"File %s yields a weird dataframe on read \n%s"%(ifile,df)
            res = {key: val for key, val in zip(df.values[row_freq + 1:, col_lab].tolist(),
                                                 df.values[row_freq + 1:, col_freq].tolist())}
    else:
        res = freq_ascii2dict(ifile)
    if defrag is not None:
        res = {defrag_key(key, defrag=defrag):val for key, val in res.items()}

    return res

def freq_ascii2dict(ifile, comment=["#"]):
    r"""
    Reads an ASCII file that contains contact frequencies
    (1st column) and contact labels (2nd and/or 3rd column).
    Columns are separated by tabs or spaces.

    Contact labels have to come after the frequency in the
    form of "res1 res2, "res1-res2" or "res1 - res2",

    Columns other than the frequencies and the residue labels are ignored.

    Examples
    --------
    File produced by mdciao:

    >>> #freq              label              residue idxs  sum
    >>> 0.59 R389@G.H5.21    - L394@G.H5.26    348 353    0.59
    >>> 0.46 L394@G.H5.26    - K270@6.32x32    353 972    1.05
    >>> 0.34 L388@G.H5.20    - L394@G.H5.26    347 353    1.39
    >>> 0.32 L394@G.H5.26    - L230@5.69x69    353 957    1.71
    >>> 0.04 R385@G.H5.17    - L394@G.H5.26    344 353    1.75

    Minimal file with mixed labeling

    >>> 1 ALA30-GLU50
    >>> .5 ASP31 - GLU51
    >>> .1 ASP31 GLU50


    TODO use pandas to allow more flex, not needed for the moment

    Parameters
    ----------
    ifile : str
        The filename to be read
    comment : list of chars
        Any line starting with any of these
        characters will be ignored
    Returns
    -------
    freqdict : dictionary
        Keys are "res1-res2" (regardless of input)
        and values are freqs

    """
    #TODO consider using pandas
    outdict = {}
    with open(ifile) as f:
        for iline in f.read().splitlines():
            if iline.strip()[0] not in comment and len(iline)>0:
                try:
                    iline = iline.replace("-"," ").split()
                    freq, names = float(iline[0]),"%s-%s"%(iline[1],iline[2])
                    outdict[names]=float(freq)
                except ValueError:
                    print(iline)
                    raise
    return outdict

_symbols =  ['alpha','beta','gamma', 'mu', "Sigma"]+["AA"]
_scripts =  ["^","_"]


def replace4latex(istr,
                  sindex=["_", "^"],
                  symbols=["alpha", "beta", "gamma", "sigma", "mu", "aa"],
                  enclose_pure_text=False
                  ):
    r"""
    Return a string where symbols and super/sub-indices have been prepared for LaTeX

    One quirk: when sub- or superindexing, the following
    types get protected in curly brackets to
    avoid only sub/super indexing the first character:

     * fully numeric: C_{300}
     * fully alphabetical: GLY_{ACE}
     * containing dots: L394^{G.H.26}

    BUT mixed \beta_2AR are left unprotected:

    >>> replace4latex("mdciao can alpha Sigma_2 beta2AR ACE_GLY GLU30^3.50 no [frag1-WT] problem!")
    'mdciao can $\\alpha$ $\\Sigma\\mathrm{_{2}}$ $\\beta\\mathrm{_2AR}$ $\\mathrm{{ACE}_{GLY}}$ $\\mathrm{GLU30^{3.50}}$ no [frag1-WT] problem!'


    Parameters
    ----------
    istr : str
        The string to be prepare for LaTeX mathmode
        If a $ sign is already in :obj:`istr`,
        nothing will happen
        If a word in :obj:`istr` contains
        the same :obj:`sindex` character more than
        once, it'll be skipped (ask [Knut](https://tex.stackexchange.com/questions/253080/why-am-i-getting-a-double-subscript-error))
    sindex : list
        The characters that indicate super- and sub-indices
    symbols : list
        The words that should be considered LaTeX symbols

    Returns
    -------
    lstr : str
        The string with LaTex-mathmode insertions
    """
    if "$" in istr:
        return istr
    pattern = " |" + "|".join(symbols)
    bits = [bit for bit in _re.split("(?i)(%s)" % pattern, istr.replace("\n", " ? ")) if len(bit) > 0]
    for ii in range(len(bits)):
        if bits[ii].lower() in symbols:
            bits[ii] = "$\%s$" % bits[ii]
        elif any([ss == bits[ii] for ss in sindex]):
            bits[ii] = "$%s$" % bits[ii]
        elif any([ss in bits[ii] for ss in sindex]):
            ibit = bits[ii]
            if any([ibit.count(ss)>1 for ss in sindex]):
                continue
            words = [word for word in _re.split("(%s)" % "|".join(["\%s" % ss for ss in sindex]), ibit)
                     if len(word) > 0]
            for ww in range(len(words)):
                word = words[ww]
                if word.isnumeric() or word.isalpha() or "." in word and word not in sindex: #Also gets 3.50
                    words[ww] = "{%s}" % word
            ibit = "".join(words)
            bits[ii] = "$\mathrm{%s}$" % ibit
        else:
            if enclose_pure_text and any([c.isalpha()  for c in bits[ii]]):
                bits[ii] = "$\mathrm{%s}$" % bits[ii]

    return "".join(bits).replace("$$", "").replace(" ? ","\n")

def _replace_regex_special_chars(word,
                                 repl_char="!",
                                 special_chars=["^", "[", "]", "(", ")"]):
    r"""
    Ad-hoc method to replace special regexp-chars with something else before
    computing char positions using regexp.finditer

    Note:
    this method only makes sense because downstream from here, finditer is used
    to search a substring in a string and special chars break that search.

    Note:
    somewhere, a dev that knows how to use regex is crying

    Parameters
    ----------
    word : str
    repl_char : char, default is '!'
        The replacement character
    special_chars : list
        The characters that trigger replacement

    Returns
    -------
    word : str
        A string with all special characters repaced with :obj:`repl_char`

    """
    for sp in special_chars:
        word = word.replace(sp, repl_char)
    return word

def latex_mathmode(istr, enclose=True):
    r"""
    Prepend *symbol* words with "\\ " and protect *non-symbol* words with '\\mathrm{}'

    * *symbol* words are things that can
      be interpreted by LaTeX in math mode, e.g.
      '\\alpha' or '\\AA'
    * *non-symbol* words are everything else

    Works "opposite" to :obj:`replace4latex` and for the moment
    it's my (very bad) solution for latexifying contact-labels' fragments
    as super indices where the fragments themselves contain
    sub-indices (GLU30^$\beta_2AR}


    >>> replace4latex("There's an alpha and a beta here, also C_200")
    "There's an $\alpha$ and a $\beta$ here, also $C_{200}$"

    >>> latex_mathmode("There's an alpha and a beta here, also C_200")
    "$\\mathrm{There's an }\\alpha\\mathrm{ and a }\\beta\\mathrm{ here, also C_200}$"

    Parameters
    ----------
    istr : string
    enclose : bool, default is True
        Return string enclosed in
        dollar-signs: '$string$'
        Use False for cases where
        the LaTeX math-mode is already
        active

    Returns
    -------
    istr : string
    """
    output = []
    exp = "(%s)" % "|".join(["\%s" % ss if ss == "^" else "%s" % ss for ss in _symbols])
    for word in _re.split(exp, istr):
        if len(word) > 0:
            if word in _symbols:
                word = "\\%s" % word
            else:
                word = "\\mathrm{%s}" % word
            output.append(word)
    output = "".join(output)
    if enclose:
        output= "$%s$"%output
    return output

def latex_superscript_fragments(contact_label, defrag="@"):
    r"""
    Format fragment descriptors as Latex math-mode superscripts

    Thinly wrap around :obj:`_latex_superscript_one_fragment` with :obj:`splitlabel`

    Parameters
    ----------
    contact_label : str
        contact label of any form,
        as long as to AAs are joined
        with '-' character
    defrag : char, default is '@'
        The character to divide
        residue and fragment label
    Returns
    -------
    contact_label : str

    """
    return '-'.join(_latex_superscript_one_fragment(w, defrag=defrag) for w in splitlabel(contact_label, "-"))

def _latex_superscript_one_fragment(label, defrag="@"):
    r"""
    Format s.t. the fragment descriptor appears as superindex in LaTeX math-mode

    Parameters
    ----------
    label : str
        Contact label, "GLU30" and
        optionally "GLU30@beta_2AR"
    defrag : char, default is '@'
        The character to divide
        residue and fragment label

    Returns
    -------
    label : str
    """
    words = label.split(defrag,maxsplit=1)
    if len(words)==1:
        return label
    elif len(words)==2:
        return words[0]+"$^{%s}$" %replace4latex(words[1], enclose_pure_text=True).replace("$","")

def _label2componentsdict(istr,sep="-",defrag="@",
                          assume_ctc_label=True):
    r"""
    Identify the components of label like 'residue1@frag1-residue2@frag2' and return them as dictionary

    Parameters
    ----------
    istr : str
        Can be of any of these forms:
        * res1
        * res1@frag1
        * res1@frag1-res2
        * res1@frag1-res2@frag2
        * res1-res2@frag2
        * res1-res2

        The fragment names can contain the separator, e.g.
        'res1@B2AR-CT-res2@Gprot' is possible, but residue
        names cannot.

        The special case 'res1@frag1-r2' is handled with
        the parameter :obj:`assume_ctc_label` (see below)

        Labels have to start with a residue.
    sep : char, default is "-"
        The character that separates pairs of labels
    defrag : char, default is "@"
        The character that separates residues form their host fragment
    assume_ctc_label : bool, default is True
        In special cases of the form 'res1@frag1-r2', assume
        this is a contact label, i.e. 'r2' does not
        belong to the name of the fragment of res1, but is
        the second residue.

    Returns
    -------
    label : dict
        A dictionary tuple with the components present in :obj:`istr`.
        Keys can be 'res1','frag1','res2','frag2'
    """
    assert len(sep)==len(defrag)==1, "The 'sep' and 'defrag' arguments have to have both len 1, have " \
                                     "instead %s (%u) %s (%u)"%(sep,len(sep),defrag,len(defrag))

    bits = {}

    if defrag not in istr:
        for ii, ires in enumerate(istr.split(sep),start=1):
            bits["res%u"%ii]=ires
    else:
        spans = [0] + _np.hstack([m.span() for m in _re.finditer(defrag, istr)]).tolist() + [len(istr)]

        # Counters
        r, f = 1, 1
        for ii, jj in _np.reshape(spans, (-1, 2)):
            iw = istr[ii:jj + 1]
            #print(iw, ii, jj)

            if sep in iw and ii == 0:
                ires, jres = iw.replace(defrag,"").split(sep)
                bits["res%u"%r]=ires
                r+=1
                bits["res%u"%r]=jres
                r+=1
                f+=1 # because we've already established res1 hasn't any fragment
            else:
                if defrag not in iw:
                    if sep not in iw or not assume_ctc_label:
                        bits["frag%u"%f]=iw
                        f+=1
                    elif sep in iw and assume_ctc_label:
                        ires, ifrag = [jw[::-1] for jw in iw[::-1].split(sep, 1)]
                        if "res1" in bits.keys():
                            if "frag1" in bits.keys():
                                bits["frag%u"%f]=iw
                                f+=1
                            else:
                                if "res2" in bits.keys():
                                    bits["frag%u" % f] = iw
                                else:
                                    bits["frag%u"%f]=ifrag
                                    f+=1
                                    bits["res%u"%r]=ires
                                    r+=1

                else:
                    assert iw.endswith(defrag)
                    if sep not in iw:
                        bits["res%u"%r]=iw.split(defrag)[0]
                        r+=1
                    else:
                        ires, ifrag = [jw[::-1] for jw in iw[::-1][1:].split(sep, 1)]
                        bits["frag%u"%f]=ifrag
                        bits["res%u"%r]=ires
                        f+=1
                        r+=1


    return bits

def splitlabel(label, sep="-", defrag="@"):
    r"""
    Split a contact label. Analogous to label.split(sep) but more robust
    because fragment names can contain the separator character.

    Parameters
    ----------
    label : str
        Can be of any of these forms:
         * res1
         * res1@frag1
         * res1@frag1-res2
         * res1@frag1-res2@frag2
         * res1-res2@frag2
         * res1-res2

        The fragment names can contain the separator, e.g.
        'res1@B2AR-CT-res2@Gprot' is possible. Residue
        names cannot contain the separator.

        The method assumes that labels start with a residue,
        (see above), else you'll get weird behaviour.
    sep : char, default is "-"
        The character that separates pairs of labels
    defrag : char, default is "@"
        The character that separates residues form their host fragment

    Returns
    -------
    split : list
        A list equivalent to having used label.split(sep)
        but the separator is ignored in the fragment labels.
    """

    bits = _label2componentsdict(label,sep=sep,defrag=defrag)

    split = [bits["res1"]]
    if "frag1" in bits.keys():
        split[0] += "%s%s" % (defrag,bits["frag1"])
    if "res2" in bits.keys():
        split.append(bits["res2"])
        if "frag2" in bits.keys():
            split[1] += "%s%s" % (defrag,bits["frag2"])
    return split

def intblocks_in_str(istr):
    r"""
    Return the integers that appear as contiguous blocks in strings

    E.g.  "GLU30@3.50-GDP396@frag1" returns [30,3,50,396,1]

    Parameters
    ----------
    istr : string

    Returns
    -------
    ints : list

    """
    intblocks = _cranges([char.isdigit() for char in istr])[True]
    return [int("".join([istr[idx] for idx in block])) for block in intblocks]

def iterate_and_inform_lambdas(ixtc,chunksize, stride=1, top=None):
    r"""
    Given a trajectory (as object or file), returns
    a strided, chunked iterator and function for progress report

    Parameters
    ----------
    ixtc: str (filename) or :obj:`mdtraj.Trajectory` object
    chunksize: int
        The trajectory will be iterated over in chunks of this many frames
    stride: int, default is 1
        The stride with which to iterate over the trajectory
    top:  str (filename) or :obj:`mdtraj.Topology`
        If :obj:`ixtc` is a filename, the topology needed to read it

    Returns
    -------

    iterate, inform

    iterate: lambda(ixtc)
        strided, chunked iterator over :obj:`ixtc`

    inform: lambda(ixtc, traj_idx, chunk_idx, running_f)
        iterator that prints out streaming progress for every iteration

    Note
    ----

    The lambdas returned differ depending on the type of input, but signature
    is the same, s.t. the user does not have to care in posterior use

    """
    if isinstance(ixtc, _md.Trajectory):
        # TODO it's time to use yield here, this is killing performance
        iterate = lambda ixtc: [ixtc[idxs] for idxs in re_warp(_np.arange(ixtc.n_frames)[::stride], chunksize)]
        inform = lambda ixtc, traj_idx, chunk_idx, running_f: \
            print("Streaming over trajectory object nr. %3u (%6u frames, %6u with stride %2u) in chunks of "
                  "%3u frames. Now at chunk nr %4u, frames so far %6u" %
                  (traj_idx, ixtc.n_frames, _np.ceil(ixtc.n_frames/stride), stride, chunksize, chunk_idx, running_f), end="\r", flush=True)
    elif ixtc.endswith(".pdb") or ixtc.endswith(".pdb.gz") or ixtc.endswith(".gro"):
        iterate =  lambda ixtc: [_md.load(ixtc)[::stride]]
        inform  =  lambda ixtc, traj_idx, chunk_idx, running_f: \
            print("Loaded %20s (nr. %3u) in full, using stride %2u but ignoring chunksize of "
                  "%6u frames. This number should always be 0 : %4u. Total frames loaded %6u" %
                  (ixtc, traj_idx, stride, chunksize, chunk_idx, running_f), end="\r", flush=True)
    else:
        iterate = lambda ixtc: _md.iterload(ixtc, top=top, stride=stride, chunk=int(_np.round(chunksize / stride)))
        inform = lambda ixtc, traj_idx, chunk_idx, running_f: \
            print("Streaming %20s (nr. %3u) with stride %2u in chunks of "
                  "%6u frames. Now at chunk nr %4u, frames so far %6u" %
                  (ixtc, traj_idx, stride, chunksize, chunk_idx, running_f), end="\r", flush=True)
    return iterate, inform

def choose_options_descencing(options,
                              fmt="%s",
                              dont_accept=["none", "na"]):
    r"""
    Return the first entry that's acceptable according to some rule

    If no is found, "" is returned
    Parameters
    ----------
    options : list
    fmt : str, default is "%s"
        You can specify a different
        format here. Will only
        apply in case something
        is returned
    dont_accept : list
        Move down the list if
        current item is one
        of these

    Returns
    -------
    best : str
        Either the best entry in :obj:`options`
        or "" if no option was found
    """
    for option in options:
        if str(option).lower() not in dont_accept:
            return fmt%str(option)
    return ""


def fnmatch_ex(patterns_as_csv, list_of_keys):
    r"""
    Match the keys in :obj:`list_of_keys` against some naming patterns
    using Unix filename pattern matching
    TODO include link:  https://docs.python.org/3/library/fnmatch.html

    This method also allows for exclusions (grep -e)

    TODO: find out if regular expression re.findall() is better

    Uses fnmatch under the hood

    Parameters
    ----------
    patterns_as_csv : str
        Patterns to include or exclude, separated by commas, e.g.
        * "H*,-H8" will include all TMs but not H8
        * "G.S*" will include all beta-sheets
    list_of_keys : list
        Keys against which to match the patterns, e.g.
        * ["H1","ICL1", "H2"..."ICL3","H6", "H7", "H8"]

    Returns
    -------
    matching_keys : list

    """
    include_patterns = [pattern for pattern in patterns_as_csv.split(",") if not pattern.startswith("-")]
    exclude_patterns = [pattern[1:] for pattern in patterns_as_csv.split(",") if pattern.startswith("-")]
    #print(include_patterns)
    #print(exclude_patterns)
    # Define the match using a lambda
    matches_include = lambda key : any([_fnmatch(str(key), patt) for patt in include_patterns])
    matches_exclude = lambda key : any([_fnmatch(str(key), patt) for patt in exclude_patterns])
    passes_filter = lambda key : matches_include(key) and not matches_exclude(key)
    outgroup = []
    for key in list_of_keys:
        #print(key, matches_include(key),matches_exclude(key),include_patterns, exclude_patterns)
        if passes_filter(key):
            outgroup.append(key)
    return outgroup

def match_dict_by_patterns(patterns_as_csv, index_dict, verbose=False):
    r"""
    Joins all the values in an input dictionary if their key matches
    some patterns. This method also allows for exclusions (grep -e)

    TODO: find out if regular expression re.findall() is better

    Parameters
    ----------
    patterns_as_csv : str
        Comma-separated patterns to include or exclude, separated by commas, e.g.
        * "H*,-H8" will include all TMs but not H8
        * "G.S*" will include all beta-sheets
    index_dict : dictionary
        It is expected to contain iterable of ints or floats or anything that
        is "joinable" via np.hstack. Typically, something like:
        * {"H1":[0,1,...30], "ICL1":[31,32,...40],...}

    Returns
    -------
    matching_keys, matching_values : list, array of joined values

    """
    matching_keys =   fnmatch_ex(patterns_as_csv, index_dict.keys())
    if verbose:
        print(', '.join(matching_keys))

    if len(matching_keys)==0:
        matching_values = []
    else:
        matching_values = _np.hstack([index_dict[key] for key in matching_keys])

    return matching_keys, matching_values

def defrag_key(key, defrag="@", sep="-"):
    r"""Remove fragment information from a contact label

    Parameters
    ----------
    key : str
        Contact label with some sort of pair information
        e.g. e.g. R1@frag1-E2@frag2->R1-E2
    defrag: char, default is "@"
        Character that indicates the beginning of the
        fragment
    sep : char, default is "-"
        Character that indicates the separation
        between first and second residue of the pair

    Returns
    -------

    """
    return sep.join([kk.split(defrag,1)[0].strip(" ") for kk in splitlabel(key,sep)])

def df_str_formatters(df):
    r"""
    Return formatters for :obj:`~pandas.DataFrame.to_string'

    In principle, this should be solved by
    https://github.com/pandas-dev/pandas/issues/13032,
    but I cannot get it to work

    Parameters
    ----------
    df : :obj:`~pandas.DataFrame`

    Returns
    -------
    formatters : dict
        Keyed with :obj:`df`-keys
        and valued with lambdas
        s.t. formatters[key][istr]=formatted_istr

    """
    formatters = {}
    for key in df.keys():
        fmt = "%%-%us"%max([len(ii)+1 for ii in df[key]])
        formatters[key]=lambda istr : fmt%istr
    return formatters

class FilenameGenerator(object):
    r"""
    Generate per project filenames when you need them

    This is a WIP to consolidate all filenaming in one place,
    s.t. all sanitizing and project-specific naming operations happen
    here and not in the cli methods

    A named tuple would've been enough, but we need some
     methods for dynamic naming (e.g. per-residue or per-traj)

    """

    def __init__(self, output_desc, ctc_cutoff_Ang, output_dir, graphic_ext, table_ext, graphic_dpi, t_unit):

        self._graphic_ext = graphic_ext.strip(".")
        self._output_desc = output_desc.strip(".")
        self._ctc_cutoff_Ang = ctc_cutoff_Ang
        self._output_dir = output_dir
        self._graphic_dpi = graphic_dpi
        self._t_unit = t_unit
        self._allowed_table_exts = ["dat", "txt", "xlsx", "ods"] #TODO what about npy?
        assert str(table_ext).lower != "none"
        self._table_ext = str(table_ext).lower().strip(".")
        if self._table_ext not in self._allowed_table_exts:
            raise ValueError("The table extension, cant be '%s', "
                             "has be one of %s"%(table_ext,self._allowed_table_exts))


    @property
    def output_dir(self):
        return self._output_dir
    @property
    def basename_wo_ext(self):
        return "%s.overall@%2.1f_Ang" % (self.output_desc,
                                         self.ctc_cutoff_Ang)
    @property
    def ctc_cutoff_Ang(self):
        return self._ctc_cutoff_Ang

    @property
    def output_desc(self):
        return self._output_desc.replace(" ","_")

    @property
    def fullpath_overall_no_ext(self):
        return _path.join(self.output_dir, self.basename_wo_ext)

    @property
    def graphic_ext(self):
        return self._graphic_ext

    @property
    def graphic_dpi(self):
        return self._graphic_dpi
    @property
    def table_ext(self):
        return self._table_ext

    @property
    def t_unit(self):
        return self._t_unit
    @property
    def fullpath_overall_fig(self):
        return ".".join([self.fullpath_overall_no_ext, self.graphic_ext])

    def fname_per_residue_table(self,istr):
        assert self.table_ext is not None
        fname = '%s.%s@%2.1f_Ang.%s' % (self.output_desc,
                                        istr.replace('*', "").replace(" ","_"),
                                        self.ctc_cutoff_Ang,
                                        self.table_ext)
        return _path.join(self.output_dir, fname)

    def fname_per_site_table(self, istr):
        return self.fname_per_residue_table(istr)


    def fname_timetrace_fig(self, surname):
        return '%s.%s.time_trace@%2.1f_Ang.%s' % (self.output_desc,
                                                  surname.replace(" ", "_"),
                                                  self.ctc_cutoff_Ang,
                                                  self.graphic_ext)
    @property
    def fullpath_overall_excel(self):
        return ".".join([self.fullpath_overall_no_ext, "xlsx"])

    @property
    def fullpath_overall_dat(self):
        return ".".join([self.fullpath_overall_no_ext, "dat"])

    @property
    def fullpath_pdb(self):
        return ".".join([self.fullpath_overall_no_ext, "as_bfactors.pdb"])

    @property
    def fullpath_matrix(self):
        return self.fullpath_overall_fig.replace("overall@", "matrix@")

    @property
    def fullpath_flare_vec(self):
        if self.graphic_ext == "svg":
            gx = self.graphic_ext
        else:
            gx = "pdf"
        return '.'.join([self.fullpath_overall_no_ext.replace("overall@", "flare@"),gx])

